#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate."""

import os

from unittest import TestCase

from debian.changelog import Changelog

from debmutate import version_string


class DebmutateVersion(TestCase):
    def test_matches_package_version(self):
        if not os.path.exists("debian/changelog"):
            self.skipTest(
                "no debian/changelog available. "
                "Running outside of source tree?"
            )
        with open("debian/changelog") as f:
            cl = Changelog(f, max_blocks=1)
        cl_version = str(cl.version)
        cl_version = cl_version.split("+")[0]
        cl_version = cl_version.split("~")[0]
        self.assertEqual(cl_version, version_string)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__all__ = [
    'TestCase',
    'TestCaseInTempDir',
    ]

import os
import tempfile
import unittest


class TestCase(unittest.TestCase):

    def overrideEnv(self, key, value):
        oldvalue = os.environ.get(key)

        def restore():
            if oldvalue is None:
                del os.environ[key]
            else:
                os.environ[key] = value

        self.addCleanup(restore)
        os.environ[key] = value


class TestCaseInTempDir(TestCase):

    def setUp(self):
        td = tempfile.TemporaryDirectory(prefix='debmutate')
        self.test_dir = td.name
        self.addCleanup(td.cleanup)
        cwd = os.getcwd()
        self.addCleanup(os.chdir, cwd)
        os.chdir(self.test_dir)

    def build_tree_contents(self, entries):
        for entry in entries:
            if entry[0].endswith('/'):
                os.mkdir(entry[0])
            else:
                with open(entry[0], 'w') as f:
                    f.write(entry[1])

    def assertFileEqual(
            self, expected_content, path, strip_trailing_whitespace=False):
        with open(path) as f:
            content = f.read()
        if strip_trailing_whitespace:
            content = content.rstrip()
            expected_content = expected_content.rstrip()
        self.assertEqual(expected_content, content)


def test_suite():
    names = [
        'changelog',
        'control',
        'copyright',
        'deb822',
        'debcargo',
        'debhelper',
        'debmutate',
        'lintian_overrides',
        'patch',
        'reformatting',
        'vcs',
        'versions',
        'watch',
        '_rules',
        ]
    module_names = [__name__ + '.test_' + name for name in names]
    loader = unittest.TestLoader()
    return loader.loadTestsFromNames(module_names)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.copyright."""

import os
import shutil
import tempfile

from unittest import TestCase

from debian.copyright import (
    FilesParagraph,
    LicenseParagraph,
    License,
    )

from debmutate.copyright import (
    NotMachineReadableError,
    CopyrightEditor,
    )


class UpdateCopyrightTests(TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.test_dir)
        self.addCleanup(os.chdir, os.getcwd())
        os.chdir(self.test_dir)
        os.mkdir('debian')

    def test_preservable_weird_spacing(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>
# A comment


Files: *
License: GPL
Copyright: 2012...
""")

        with CopyrightEditor() as updater:
            updater.copyright.header.upstream_name = 'llintian-brush'

    def test_old_style(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
This package was debianized in 1995 by Joe Example <joe@example.com>

It was downloaded from ftp://ftp.example.com/pub/blah.
""")

        def dummy():
            with CopyrightEditor():
                pass
        self.assertRaises(NotMachineReadableError, dummy)

    def test_modify(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

Files: *
License: GPL
Copyright: 2012...
""")

        with CopyrightEditor() as updater:
            p = FilesParagraph.create(
                ['foo.c'], "2012 Joe Example", License("Apache"))
            updater.copyright.add_files_paragraph(p)
        self.assertTrue(updater.changed)
        with open('debian/copyright') as f:
            self.assertEqual("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

Files: *
License: GPL
Copyright: 2012...

Files: foo.c
Copyright: 2012 Joe Example
License: Apache
""", f.read())

    def test_add_paragraph(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>
""")

        with CopyrightEditor() as updater:
            updater.copyright.add_license_paragraph(LicenseParagraph.create(
                License("Blah", 'Blah\nblah blah\nblah\n\n')))
        self.assertTrue(updater.changed)
        with open('debian/copyright') as f:
            self.assertEqual("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

License: Blah
 Blah
 blah blah
 blah
 .
""", f.read())

    def test_preserve_whitespace(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

License: Blah
 blah
 .
""")

        with CopyrightEditor() as updater:
            license_para = list(updater.copyright.all_license_paragraphs())[0]
            self.assertEqual(
                'License: Blah\n blah\n .\n',
                license_para.dump())
            self.assertEqual("blah\n", license_para.license.text)

    def test_remove_paragraph(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

License: Blah
 blah
 .
""")

        with CopyrightEditor() as updater:
            license_para = list(updater.copyright.all_license_paragraphs())[0]
            updater.remove(license_para)
        with open('debian/copyright') as f:
            self.assertEqual("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>
""", f.read())

    def test_append_paragraph(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

""")

        with CopyrightEditor() as updater:
            p = FilesParagraph.create(
                ['foo.c'], "2012 Joe Example", License("Apache"))
            updater.append(p)
            p = FilesParagraph.create(
                ['bar.c'], "2013 Joe Example", License("Apache"))
            updater.append(p)
        with open('debian/copyright') as f:
            self.assertEqual("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

Files: foo.c
Copyright: 2012 Joe Example
License: Apache

Files: bar.c
Copyright: 2013 Joe Example
License: Apache
""", f.read())

    def test_insert_paragraph(self):
        with open('debian/copyright', 'w') as f:
            f.write("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

Files: bar.c
Copyright: 2013 Joe Example
License: Apache
""")

        with CopyrightEditor() as updater:
            p = FilesParagraph.create(
                ['foo.c'], "2012 Joe Example", License("Apache"))
            updater.insert(0, p)
        with open('debian/copyright') as f:
            self.assertEqual("""\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: lintian-brush
Upstream-Contact: Jelmer <jelmer@samba.org>

Files: foo.c
Copyright: 2012 Joe Example
License: Apache

Files: bar.c
Copyright: 2013 Joe Example
License: Apache
""", f.read())



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.rules."""

from . import TestCase, TestCaseInTempDir

from debmutate._rules import (
    Makefile,
    Rule,
    dh_invoke_drop_with,
    dh_invoke_add_with,
    dh_invoke_get_with,
    discard_pointless_override,
    matches_wildcard,
    update_rules,
    )


class MakefileParseTests(TestCase):

    def test_simple_rule(self):
        mf = Makefile.from_bytes(b"""\
all:
\ttest
""")
        self.assertEqual(mf.contents, [Rule(b'all', commands=[b'test'])])

    def test_conditional(self):
        mf = Makefile.from_bytes(b"""\
all:
ifeq (foo, bar)
\ttest
  endif
""")
        self.assertEqual(1, len(mf.contents))

    def test_conditional_rule(self):
        mf = Makefile.from_bytes(b"""\
all: %: test
\ttest
""")
        self.assertEqual(
            mf.contents,
            [Rule(b'all', commands=[b'test'], prereq_targets=[b'%:', b'test'])]
            )

    def test_rule_with_comment(self):
        mf = Makefile.from_bytes(b"""\
rule1:
\ttest1

# And this is a comment
rule2:
\ttest2
""")
        self.assertEqual(
            mf.contents,
            [Rule(b'rule1', commands=[b'test1']),
             b'',
             Rule(b'rule2', commands=[b'test2'],
                  precomment=[b'# And this is a comment'])
             ])


class UpdateRulesTests(TestCaseInTempDir):

    def test_update_command(self):
        self.build_tree_contents([('debian/', ), ('debian/rules', """\
SOMETHING = 1

all:
\techo blah
\techo foo
""")])

        def replace(line, target):
            if line == b'echo blah':
                return b'echo bloe'
            return line
        self.assertTrue(update_rules(replace))
        self.assertFalse(update_rules(replace))
        self.assertFileEqual("""\
SOMETHING = 1

all:
\techo bloe
\techo foo
""", 'debian/rules')

    def test_continuation(self):
        self.build_tree_contents([('debian/', ), ('debian/rules', """\
SOMETHING = 1

all:
\techo blah \\
foo
""")])

        def replace(line, target):
            if line == b'echo blah \\\nfoo':
                return b'echo bloe'
            return line
        self.assertTrue(update_rules(replace))
        self.assertFalse(update_rules(replace))
        self.assertFileEqual("""\
SOMETHING = 1

all:
\techo bloe
""", 'debian/rules')

    def test_empty_line(self):
        self.build_tree_contents([('debian/', ), ('debian/rules', """\
SOMETHING = 1

all:
\techo blah

\techo foo
""")])

        def replace(line, target):
            if line == b'echo foo':
                return b'echo bloe'
            return line
        self.assertTrue(update_rules(replace))
        self.assertFalse(update_rules(replace))
        self.assertFileEqual("""\
SOMETHING = 1

all:
\techo blah

\techo bloe
""", 'debian/rules')

    def test_remove_variable(self):
        self.build_tree_contents([('debian/', ), ('debian/rules', """\
# Comment
SOMETHING = 1

all:
\techo blah
""")])

        def remove(line):
            if line == b'SOMETHING = 1':
                return None
            return line
        self.assertTrue(update_rules(global_line_cb=remove))
        self.assertFalse(update_rules(global_line_cb=remove))
        self.assertFileEqual("""\
# Comment

all:
\techo blah
""", 'debian/rules')

    def test_remove_variable_with_comment(self):
        self.build_tree_contents([('debian/', ), ('debian/rules', """\
# Comment
SOMETHING = 1

all:
\techo blah
""")])

        def remove(line):
            if line == b'SOMETHING = 1':
                return None
            return line
        self.assertTrue(update_rules(
            global_line_cb=remove, drop_related_comments=True))
        self.assertFalse(update_rules(global_line_cb=remove))
        self.assertFileEqual("""\

all:
\techo blah
""", 'debian/rules')

    def test_keep_rule_cb(self):
        self.build_tree_contents([('debian/', ), ('debian/rules', """\
SOMETHING = 1

all:
\techo blah

none:
\techo foo
""")])

        def discard_none(rule):
            if rule.target == b'none':
                rule.clear()
        self.assertTrue(update_rules(rule_cb=discard_none))
        self.assertFalse(update_rules(rule_cb=discard_none))
        self.assertFileEqual("""\
SOMETHING = 1

all:
\techo blah
""", 'debian/rules')


class MakefileTests(TestCase):

    def test_add_rule(self):
        mf = Makefile.from_bytes(b"""\
SOMETHING = 1

all:
\techo blah

# Original rule
# Multi-line comment
none:
\techo foo
""")
        r = mf.add_rule(b'blah', precomment=[b'# A new rule'])
        self.assertIsInstance(r, Rule)
        r.append_command(b'echo really blah')
        self.assertEqual(b"""\
SOMETHING = 1

all:
\techo blah

# Original rule
# Multi-line comment
none:
\techo foo

# A new rule
blah:
\techo really blah
""", mf.dump())

    def test_get_variable(self):
        mf = Makefile.from_bytes(b"""\
SOMETHING = 1
export SOMETHING_ELSE = 2
SOMETHING_EXPORTED := 4

all:
\techo blah

""")
        self.assertEqual(b'1', mf.get_variable(b'SOMETHING'))
        self.assertEqual(b'2', mf.get_variable(b'SOMETHING_ELSE'))
        self.assertEqual(b'4', mf.get_variable(b'SOMETHING_EXPORTED'))
        self.assertRaises(KeyError, mf.get_variable, b'SOMETHING_MISSING')


class InvokeDropWithTests(TestCase):

    def test_drop_with(self):
        self.assertEqual(
            b'dh',
            dh_invoke_drop_with(b'dh --with=blah', b'blah'))
        self.assertEqual(
            b'dh --with=foo',
            dh_invoke_drop_with(b'dh --with=blah,foo', b'blah'))
        self.assertEqual(
            b'dh --with=foo --other',
            dh_invoke_drop_with(b'dh --with=blah,foo --other', b'blah'))
        self.assertEqual(
            b'dh',
            dh_invoke_drop_with(b'dh --with=blah', b'blah'))
        self.assertEqual(
            b'dh --with=foo',
            dh_invoke_drop_with(b'dh --with=foo,blah', b'blah'))
        self.assertEqual(
            b'dh $@ --verbose --with autoreconf,cme-upgrade',
            dh_invoke_drop_with(
                b'dh $@ --verbose --with autoreconf,systemd,cme-upgrade',
                b'systemd'))
        self.assertEqual(
            b'dh $@ --with gir,python3,sphinxdoc --without autoreconf '
            b'--buildsystem=cmake',
            dh_invoke_drop_with(
                b'dh $@ --with gir,python3,sphinxdoc,systemd '
                b'--without autoreconf --buildsystem=cmake',
                b'systemd'))
        self.assertEqual(
            b'dh $@',
            dh_invoke_drop_with(
                b'dh $@ --with systemd',
                b'systemd'))


class DhInvokeGetWithsTests(TestCase):

    def test_simple(self):
        self.assertEqual(
            ['blah'],
            dh_invoke_get_with(b'dh --with=blah --foo'))
        self.assertEqual(
            ['blah'],
            dh_invoke_get_with(b'dh --with=blah'))
        self.assertEqual(
            ['blah', 'blie'],
            dh_invoke_get_with(b'dh --with=blah --with blie'))
        self.assertEqual(
            ['blah', 'blie'],
            dh_invoke_get_with(b'dh --with=blah,blie'))


class InvokeAddWithTests(TestCase):

    def test_add_with(self):
        self.assertEqual(
            b'dh --with=blah',
            dh_invoke_add_with(b'dh', b'blah'))
        self.assertEqual(
            b'dh --with=foo,blah',
            dh_invoke_add_with(b'dh --with=foo', b'blah'))
        self.assertEqual(
            b'dh --with=foo,blah --other',
            dh_invoke_add_with(b'dh --with=foo --other', b'blah'))


class MatchesWildcardTests(TestCase):

    def test_some(self):
        self.assertTrue(matches_wildcard('foo', 'foo'))
        self.assertTrue(matches_wildcard('foo', 'fo%'))
        self.assertTrue(matches_wildcard('foo', '%'))
        self.assertFalse(matches_wildcard('foo', 'bar'))
        self.assertFalse(matches_wildcard('foo', 'fo'))
        self.assertFalse(matches_wildcard('foo', 'oo'))
        self.assertFalse(matches_wildcard('foo', 'b%'))


class DiscardPointlessOverrideTests(TestCase):

    def test_simple(self):
        mf = Makefile.from_bytes(b"""\
override_dh_blah:
\tdh_blah

.PHONY: override_dh_blah
""")

        rule = next(mf.iter_rules(b'override_dh_blah'))
        discard_pointless_override(mf, rule)
        self.assertEqual(rule.lines, [])
        phony = next(mf.iter_rules(b'.PHONY'))
        self.assertEqual([], phony.components)



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with debcargo files."""

from collections.abc import MutableMapping
from itertools import chain
import os
import re
from typing import Optional, Tuple

from debian.changelog import Changelog

from tomlkit import loads, dumps

from .reformatting import Editor


DEFAULT_MAINTAINER = (
    "Debian Rust Maintainers <pkg-rust-maintainers@alioth-lists.debian.net>")
DEFAULT_SECTION = 'rust'
CURRENT_STANDARDS_VERSION = '4.5.1'


class AutomaticFieldUnknown(KeyError):
    """Field is generated automatically, and value can not be determined."""


def semver_pair(version):
    import semver
    parsed = semver.VersionInfo.parse(version)
    return '%d.%d' % (parsed.major, parsed.minor)


class TomlEditor(Editor):

    def _parse(self, content):
        return loads(content)

    def _format(self, parsed):
        if not parsed:
            return None
        return dumps(parsed)

    def __contains__(self, key):
        return key in self._parsed

    def __getitem__(self, key):
        return self._parsed[key]

    def __delitem__(self, key):
        del self._parsed[key]

    def __setitem__(self, key, value):
        self._parsed[key] = value

    def get(self, key, default=None):
        return self._parsed.get(key, default)


class DebcargoEditor(TomlEditor):

    def __init__(
            self, path: str = 'debian/debcargo.toml',
            allow_reformatting: Optional[bool] = None,
            allow_missing: bool = False):
        super().__init__(
            path=path, allow_reformatting=allow_reformatting)
        self.allow_missing = allow_missing

    def __repr__(self):
        return "{}({!r}, allow_reformatting={!r}, allow_missing={!r})".format(
            type(self).__name__, self.path, self.allow_reformatting,
            self.allow_missing)

    def _nonexistant(self):
        if self.allow_missing:
            return {}
        raise


class ShimParagraph(MutableMapping):

    def items(self):
        for key in iter(self):
            try:
                yield key, self[key]
            except AutomaticFieldUnknown:
                pass


class DebcargoSourceShimEditor(ShimParagraph):

    def __init__(self, debcargo, crate_name=None, crate_version=None,
                 cargo=None):
        self._debcargo = debcargo
        if crate_name is None:
            crate_name = cargo["package"]["name"]
        if crate_version is None:
            crate_version = cargo["package"]["version"]
        self.crate_version = crate_version
        self.crate_name = crate_name
        self.cargo = cargo

    def __getitem__(self, name):
        if name in self.SOURCE_KEY_MAP:
            (toml_name, default) = self.SOURCE_KEY_MAP[name]
            try:
                value = self._debcargo['source'][toml_name]
            except KeyError:
                if callable(default):
                    default = default(self)
                if default is not None:
                    return default
                raise KeyError(name)
            else:
                if isinstance(value, list):
                    return ', '.join(value)
                return value
        elif name in self.KEY_MAP:
            (toml_name, default) = self.KEY_MAP[name]
            try:
                value = self._debcargo[toml_name]
            except KeyError:
                if callable(default):
                    default = default(self)
                if default is not None:
                    return default
                raise KeyError(name)
            else:
                if isinstance(value, list):
                    return ', '.join(value)
                return value
        elif name == 'Source':
            if self._debcargo.get("semver_suffix", False):
                return 'rust-{}-{}'.format(
                    self.crate_name.replace('_', '-'),
                    semver_pair(self.crate_version))
            return 'rust-%s' % self.crate_name.replace('_', '-')
        elif name == 'Priority':
            return 'optional'
        elif name == 'Rules-Requires-Root':
            return 'no'
        else:
            raise KeyError(name)

    def _default_vcs_git(self):
        return ('https://salsa.debian.org/rust-team/debcargo-conf.git '
                '[src/%s]' % self.crate_name)

    def _default_vcs_browser(self):
        return ('https://salsa.debian.org/rust-team/debcargo-conf/tree/'
                'master/src/%s' % self.crate_name)

    def _build_depends(self):
        # TODO(jelmer): read Cargo.toml
        raise AutomaticFieldUnknown('Build-Depends')

    def _default_homepage(self):
        if self.cargo:
            return self.cargo["package"]["homepage"]
        else:
            raise AutomaticFieldUnknown("homepage")

    def __setitem__(self, name, value):
        if name in self.SOURCE_KEY_MAP:
            toml_name, default = self.SOURCE_KEY_MAP[name]
            if callable(default):
                default = default(self)
            if value == default:
                try:
                    del self._debcargo['source'][toml_name]
                except KeyError:
                    pass
            else:
                if 'source' not in self._debcargo:
                    self._debcargo['source'] = {}
                self._debcargo['source'][toml_name] = value
        elif name in self.KEY_MAP:
            toml_name, default = self.KEY_MAP[name]
            if callable(default):
                default = default(self)
            if value == default:
                del self._debcargo[toml_name]
            else:
                self._debcargo[toml_name] = value
        else:
            raise KeyError(name)

    def __delitem__(self, name):
        if name in self.SOURCE_KEY_MAP:
            toml_name, default = self.SOURCE_KEY_MAP[name]
            if default is None:
                try:
                    del self._debcargo['source'][toml_name]
                except KeyError:
                    pass
            else:
                raise KeyError(name)
        elif name in self.KEY_MAP:
            toml_name, default = self.KEY_MAP[name]
            if default is None:
                del self._debcargo[toml_name]
            else:
                raise KeyError(name)
        else:
            raise KeyError(name)

    def __iter__(self):
        for name in chain(self.KEY_MAP, self.SOURCE_KEY_MAP,
                          ['Source', 'Priority', 'Rules-Requires-Root']):
            try:
                self[name]
            except KeyError:
                pass
            except AutomaticFieldUnknown:
                yield name
            else:
                yield name

    def __len__(self):
        return len(self.KEY_MAP) + len(self.SOURCE_KEY_MAP) + 2

    SOURCE_KEY_MAP = {
        'Standards-Version': ('policy', CURRENT_STANDARDS_VERSION),
        'Homepage': ('homepage', _default_homepage),
        'Vcs-Git': ('vcs_git', _default_vcs_git),
        'Vcs-Browser': (
            'vcs_browser', _default_vcs_browser),
        'Section': ('section', DEFAULT_SECTION),
        'Build-Depends': ('build_depends', _build_depends),
        }
    KEY_MAP = {
        'Maintainer': ('maintainer', DEFAULT_MAINTAINER),
        'Uploaders': ('uploaders', None),
        }


class DebcargoBinaryShimEditor(ShimParagraph):

    def _provides(self):
        import semver
        parsed = semver.VersionInfo.parse(self.crate_version)
        semver_suffix = self._debcargo.get('semver_suffix', False)
        ret = []
        suffixes = []
        if not semver_suffix:
            suffixes.append('')
        suffixes.append('-%d' % (parsed.major, ))
        suffixes.append('-%d.%d' % (parsed.major, parsed.minor))
        suffixes.append(
            '-%d.%d.%d' % (parsed.major, parsed.minor, parsed.patch))
        for ver_suffix in suffixes:
            for feature in chain([None], self.features or []):
                ret.append(debcargo_binary_name(
                    self.crate_name,
                    suffix=ver_suffix + (('+' + feature) if feature else '')))
        if self.package_name in ret:
            ret.remove(self.package_name)
        if not ret:
            return None
        return '\n ' + ',\n '.join(
            ['%s (= ${binary:Version})' % p for p in ret])

    def _description(self):
        return (
            self._debcargo['description'] + ' - Rust source code\n'
            ' This package contains the source for the Rust mio crate, '
            'packaged by debcargo\n'
            'for use with cargo and dh-cargo.')

    BINARY_KEY_MAP = {
        'Section': ('section', None),
        'Depends': ('depends', None),
        'Recommends': ('recommends', None),
        'Suggests': ('suggests', None),
        'Provides': ('provides', _provides),
        'Description': ('description', _description),
        }

    def __init__(self, crate_name, crate_version, debcargo, key, package_name,
                 features):
        self.crate_name = crate_name
        self.crate_version = crate_version
        self._debcargo = debcargo
        self._key = key
        self.package_name = package_name
        self.features = features

    @property
    def _section(self):
        return 'packages.' + self._key

    def __getitem__(self, name):
        if name in self.BINARY_KEY_MAP:
            (toml_name, default) = self.BINARY_KEY_MAP[name]
            try:
                return self._debcargo[self._section][toml_name]
            except KeyError:
                if callable(default):
                    default = default(self)
                if default is None:
                    raise KeyError(name)
                return default
        elif name == 'Package':
            return self.package_name
        elif name == 'Architecture':
            return 'any'
        elif name == 'Multi-Arch':
            return 'same'
        else:
            raise KeyError(name)

    def __setitem__(self, name, value):
        if name in self.BINARY_KEY_MAP:
            (toml_name, default) = self.BINARY_KEY_MAP[name]
            if self._section not in self._debcargo:
                self._debcargo[self._section] = {}
            self._debcargo[self._section][toml_name] = value
        else:
            raise KeyError(name)

    def __delitem__(self, name):
        if name in self.BINARY_KEY_MAP:
            (toml_name, default) = self.BINARY_KEY_MAP[name]
            if default is None:
                try:
                    del self._debcargo[self._section][toml_name]
                except KeyError:
                    pass
            else:
                raise KeyError(name)
        else:
            raise KeyError(name)

    def __iter__(self):
        for name in chain(
                self.BINARY_KEY_MAP,
                ['Package', 'Provides', 'Multi-Arch', 'Architecture']):
            try:
                self[name]
            except AutomaticFieldUnknown:
                yield name
            except KeyError:
                pass
            else:
                yield name

    def __len__(self):
        return len(self.BINARY_KEY_MAP) + 1


def debcargo_version_to_semver(version):
    m = re.fullmatch('(.*)~([a-z]+)(.*)', version)
    if m:
        return '{}-{}{}'.format(m.group(1), m.group(2), m.group(3))
    return version


class DebcargoControlShimEditor:
    """Shim for debian/control that edits debian/debcargo.toml."""

    def __init__(self, debcargo_editor, crate_name, crate_version, cargo=None,
                 features=None):
        self.debcargo_editor = debcargo_editor
        self.cargo = cargo
        self.crate_name = crate_name
        self.crate_version = crate_version
        self.features = features

    def __repr__(self):
        return "{}({!r}, {!r}, {!r})".format(
            type(self).__name__, self.debcargo_editor,
            self.crate_name, self.crate_version)

    @property
    def source(self):
        return DebcargoSourceShimEditor(
            self.debcargo_editor, crate_name=self.crate_name,
            crate_version=self.crate_version, cargo=self.cargo)

    def wrap_and_sort(self, short_indent: bool = False,
                      trailing_comma: bool = False,
                      wrap_always: bool = False,
                      max_line_length: int = 79) -> None:
        pass

    def sort_binary_packages(self, keep_first: bool = False):
        pass

    @classmethod
    def from_debian_dir(cls, path, crate_name=None, crate_version=None,
                        features=None, cargo=None):
        editor = DebcargoEditor(
                os.path.join(path, 'debcargo.toml'), allow_missing=True)
        cargo_path = os.path.join(path, '..', 'Cargo.toml')
        if cargo is None:
            try:
                with open(cargo_path) as f:
                    cargo = loads(f.read())
                    crate_name = cargo["package"]["name"]
                    crate_version = cargo["package"]["version"]
                    features = list(cargo['features'])
            except FileNotFoundError:
                pass
        try:
            with open(os.path.join(path, 'changelog')) as f:
                cl = Changelog(f)
                package = cl.package
                if crate_name is None or crate_version is None:
                    crate_version = debcargo_version_to_semver(
                        cl.version.upstream_version)
                    with editor:
                        semver_suffix = editor.get("semver_suffix", False)
                    crate_name, crate_semver_version = (
                        parse_debcargo_source_name(package, semver_suffix))
        except FileNotFoundError:
            pass
        return cls(
            editor, crate_name=crate_name, crate_version=crate_version,
            cargo=cargo, features=features)

    def __enter__(self):
        self.debcargo_editor.__enter__()
        return self

    def __exit__(self, exc_typ, exc_val, exc_tb):
        self.debcargo_editor.__exit__(exc_typ, exc_val, exc_tb)
        return False

    @property
    def binaries(self):
        semver_suffix = self.debcargo_editor.get('semver_suffix', False)
        ret = [DebcargoBinaryShimEditor(
            self.crate_name, self.crate_version, self.debcargo_editor, 'lib',
            debcargo_binary_name(
                self.crate_name, '-' + semver_pair(self.crate_version)
                if semver_suffix else ''),
            self.features)]
        try:
            need_bin_package = self.debcargo_editor['bin']
        except KeyError:
            try:
                need_bin_package = not semver_suffix
            except KeyError:
                need_bin_package = False
        if need_bin_package:
            try:
                bin_name = self.debcargo_editor['bin_name']
            except KeyError:
                bin_name = self.crate_name
            ret.append(DebcargoBinaryShimEditor(
                self.crate_name, self.crate_version, self.debcargo_editor,
                'bin', bin_name, self.features))
        # TODO(jelmer): Add lib+feature packages
        return ret

    @property
    def paragraphs(self):
        return [self.source] + self.binaries


def parse_debcargo_source_name(
        source_name: str,
        semver_suffix: bool = False) -> Tuple[str, Optional[str]]:
    """Parse a debcargo source name and return crate.

    Args:
      source_name: Source package name
      semver_suffix: Whether semver_suffix is enabled
    Returns:
      tuple with crate name and optional semver
    """
    if not source_name.startswith('rust-'):
        raise ValueError(source_name)
    crate = source_name[len('rust-'):]
    crate_semver_version = None
    if semver_suffix and '-' in crate:
        crate, crate_semver_version = crate.rsplit('-', 1)
    return crate, crate_semver_version


def cargo_translate_dashes(crate):
    import subprocess
    output = subprocess.check_output(['cargo', 'search', crate])
    for line in output.splitlines(False):
        name = line.split(b' = ')[0].decode()
        return name
    return crate


def unmangle_debcargo_version(version):
    return version.replace('~', '-')


def debcargo_binary_name(crate_name, suffix=''):
    return 'librust-{}{}-dev'.format(crate_name.replace('_', '-'), suffix)


if __name__ == '__main__':
    from debian.deb822 import Deb822
    with DebcargoControlShimEditor.from_debian_dir('debian') as editor:
        print(Deb822(editor.source))
        for binary in editor.binaries:
            print(Deb822(binary))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with Debian Vcs URLs of various types."""

__all__ = [
    'split_vcs_url',
    'unsplit_vcs_url',
    'get_vcs_info',
    'mangle_version_for_git',
    'source_package_vcs',
    'gbp_expand_tag_name',
]

import re
from typing import Optional, Tuple, NamedTuple


class VcsUrl(NamedTuple):

    url: str
    branch: Optional[str]
    path: Optional[str]

    def __str__(self) -> str:
        return unsplit_vcs_url(self.url, self.branch, self.path)


def split_vcs_url(url: str) -> VcsUrl:
    """Split a Debian VCS URL.

    Args:
      url: Url to split
    Returns:
      tuple with (url, optional branch, optional path)
    """
    subpath: Optional[str]
    branch: Optional[str]
    m = re.search(r' \[([^] ]+)\]', url)
    if m:
        url = url[:m.start()] + url[m.end():]
        subpath = m.group(1)
    else:
        subpath = None
    try:
        (repo_url, branch) = url.split(' -b ', 1)
    except ValueError:
        branch = None
        repo_url = url
    return VcsUrl(repo_url, branch, subpath)


def unsplit_vcs_url(repo_url: str,
                    branch: Optional[str] = None,
                    subpath: Optional[str] = None) -> str:
    """Unsplit a Debian VCS URL.

    Args:
      repo_url: Repository URL
      branch: Branch name
      subpath: Subpath in the tree
    Returns: full URL
    """
    url = repo_url
    if branch:
        url = '{} -b {}'.format(url, branch)
    if subpath:
        url = '{} [{}]'.format(url, subpath)
    return url


def get_vcs_info(control) -> Tuple[
        Optional[str], Optional[str], Optional[str]]:
    if "Vcs-Git" in control:
        repo_url, branch, subpath = split_vcs_url(control["Vcs-Git"])
        return ("Git", repo_url, subpath)

    if "Vcs-Bzr" in control:
        return ("Bzr", control["Vcs-Bzr"], None)

    if "Vcs-Svn" in control:
        return ("Svn", control["Vcs-Svn"], None)

    if "Vcs-Hg" in control:
        repo_url, branch, subpath = split_vcs_url(control["Vcs-Hg"])
        return ("Hg", repo_url, subpath)

    return None, None, None


def mangle_version_for_git(version: str) -> str:
    """Mangle a version string for use in a Git tag.

    Args:
      version: version string to manipulate
    Returns: tag name
    """
    # See https://dep-team.pages.debian.net/deps/dep14/
    manipulated = (
        version.replace("~", "_").replace(':', '%').replace('..', '.#.'))
    if manipulated.endswith('.'):
        manipulated += '#'
    if manipulated.endswith('.lock'):
        manipulated = manipulated[:-4] + '#lock'
    return manipulated


def source_package_vcs(control) -> Tuple[str, str]:
    """Extract the Vcs URL from a source package.

    Args:
      control: A source control paragraph
    Returns:
      Tuple with Vcs type and Vcs URL
    Raises:
      KeyError: When no Vcs header was found
    """
    for prefix in ['Vcs-', 'XS-Vcs-', 'X-Vcs']:
        for field, value in control.items():
            if field.startswith(prefix):
                vcs_type = field[len(prefix):]
                if vcs_type == 'Browser':
                    continue
                return vcs_type, value
    raise KeyError


class GbpTagFormatError(Exception):
    """Unknown variable in gbp tag name."""

    def __init__(self, tag_name, variable):
        super().__init__(tag_name, variable)
        self.variable = variable
        self.tag_name = tag_name


def gbp_expand_tag_name(tag_format: str, version: str) -> str:
    # See gbp/pkg/pkgpolicy.py in gbp-buildpackage
    version_mangle_re = (
        r'%\(version'
        r'%(?P<M>[^%])'
        r'%(?P<R>([^%]|\\%))+'
        r'\)s')

    ret = tag_format
    m = re.search(version_mangle_re, tag_format)
    if m:
        ret = re.sub(version_mangle_re, "%(version)s", tag_format)
        version = version.replace(
            m.group('M'), m.group('R').replace(r'\%', '%'))

    vars = {
        'version': version,
        'hversion': version.replace('.', '-'),
    }
    try:
        return ret % vars
    except KeyError as e:
        raise GbpTagFormatError(tag_format, e.args[0])



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.control."""

import os

from . import (
    TestCase,
    TestCaseInTempDir,
    )

from debmutate.control import (
    _cdbs_resolve_conflict,
    add_dependency,
    drop_dependency,
    ensure_exact_version,
    ensure_minimum_version,
    ensure_some_version,
    ensure_relation,
    get_relation,
    guess_template_type,
    iter_relations,
    is_relation_implied,
    is_dep_implied,
    update_control,
    PkgRelation,
    format_relations,
    parse_relations,
    delete_from_list,
    ControlEditor,
    parse_standards_version,
    MissingSourceParagraph,
    )
from debmutate.reformatting import (
    GeneratedFile,
    FormattingUnpreservable,
    )


class UpdateControlTests(TestCaseInTempDir):

    def test_do_not_edit(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
# DO NOT EDIT
# This file was generated by blah

Source: blah
Testsuite: autopkgtest

""")])

        def source_cb(c):
            c['Source'] = 'blah1'
        self.assertRaises(
            GeneratedFile, update_control, source_package_cb=source_cb)

    def test_add_binary(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest

Package: blah
Description: Some description
 And there are more lines
 And more lines
""")])
        with ControlEditor('debian/control') as editor:
            editor.add_binary(
                {'Package': 'foo', 'Description': 'A new package foo'})
            self.assertEqual(
                [b['Package'] for b in editor.binaries], ['blah', 'foo'])
        self.assertTrue(editor.changed)
        self.assertEqual(editor.changed_files, ['debian/control'])

    def test_list_binaries(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest

Package: blah
Description: Some description
 And there are more lines
 And more lines
""")])
        with ControlEditor('debian/control') as editor:
            self.assertEqual(list(editor.binaries)[0]['Package'], 'blah')
        self.assertFalse(editor.changed)
        self.assertEqual(editor.changed_files, [])

    def test_no_source(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Package: blah
Testsuite: autopkgtest

Package: bar
""")])
        with ControlEditor('debian/control') as editor:
            with self.assertRaises(MissingSourceParagraph):
                editor.source

    def test_create(self):
        self.build_tree_contents([('debian/', )])
        with ControlEditor.create('debian/control') as editor:
            editor.source['Source'] = 'foo'
        self.assertFileEqual("""\
Source: foo
""", 'debian/control')

    def test_do_not_edit_no_change(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
# DO NOT EDIT
# This file was generated by blah

Source: blah
Testsuite: autopkgtest

""")])
        update_control()

    def test_unpreservable(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
# A comment
Testsuite: autopkgtest

""")])

        def update_source(control):
            control["NewField"] = "New Field"
        update_control(source_package_cb=update_source)

    def test_merge3(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest

Package: blah
Description: Some description
 And there are more lines
 And more lines
# A comment
Multi-Arch: foreign
""")])

        def update_source(control):
            control["NewField"] = "New Field"

        try:
            import merge3  # noqa: F401
        except ModuleNotFoundError:
            has_merge3 = False
        else:
            has_merge3 = (merge3.__version__ >= (0, 0, 7))
        if has_merge3:
            update_control(source_package_cb=update_source)
            self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
NewField: New Field

Package: blah
Description: Some description
 And there are more lines
 And more lines
# A comment
Multi-Arch: foreign
""", 'debian/control')
        else:
            self.assertRaises(
                FormattingUnpreservable, update_control,
                source_package_cb=update_source)

    def test_modify_source(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest
""")])

        def add_header(control):
            control["XS-Vcs-Git"] = "git://github.com/example/example"
        self.assertTrue(update_control(source_package_cb=add_header))
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
XS-Vcs-Git: git://github.com/example/example
""", 'debian/control')

    def test_wrap_and_sort(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest
Depends: package3, package2

Package: libblah
Section: extra
""")])
        with ControlEditor() as editor:
            try:
                editor.wrap_and_sort(trailing_comma=True)
            except NotImplementedError as e:
                self.skipTest(str(e))

        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
Depends: package2, package3,

Package: libblah
Section: extra
""", 'debian/control')

    def test_sort_binaries(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest
Depends: package3, package2

Package: libfoo
Section: web

Package: libblah
Section: extra
""")])
        with ControlEditor() as editor:
            editor.sort_binary_packages()

        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
Depends: package3, package2

Package: libblah
Section: extra

Package: libfoo
Section: web
""", 'debian/control')

    def test_modify_binary(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest

Package: libblah
Section: extra
""")])

        def add_header(control):
            control["Arch"] = "all"
        self.assertTrue(update_control(binary_package_cb=add_header))
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest

Package: libblah
Section: extra
Arch: all
""", 'debian/control')

    def test_doesnt_strip_whitespace(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest

""")])
        self.assertFalse(update_control())
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest

""", 'debian/control')

    def test_update_template(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
# DO NOT EDIT
# This file was generated by blah

Source: blah
Testsuite: autopkgtest
Uploaders: Jelmer Vernooij <jelmer@jelmer.uk>

"""), ('debian/control.in', """\
Source: blah
Testsuite: autopkgtest
Uploaders: @lintian-brush-test@

"""), ('debian/rules', """\
#!/usr/bin/make -f

debian/control: debian/control.in
\tsed -e 's/@lintian-brush-test@/testvalue/' < $< > $@
""")])
        os.chmod('debian/rules', 0o755)

        with ControlEditor() as updater:
            updater.source['Testsuite'] = 'autopkgtest8'
            updater.changes()
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest8
Uploaders: @lintian-brush-test@
""", "debian/control.in", strip_trailing_whitespace=True)
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest8
Uploaders: testvalue
""", "debian/control", strip_trailing_whitespace=True)
        self.assertTrue(updater.changed)
        self.assertEqual(['debian/control.in', 'debian/control'],
                         updater.changed_files)

    def test_update_template_only(self):
        self.build_tree_contents([('debian/', ), ('debian/control.in', """\
Source: blah
Testsuite: autopkgtest
Uploaders: @lintian-brush-test@

"""), ('debian/rules', """\
#!/usr/bin/make -f

debian/control: debian/control.in
\tsed -e 's/@lintian-brush-test@/testvalue/' < $< > $@
""")])
        os.chmod('debian/rules', 0o755)

        with ControlEditor() as updater:
            updater.source['Testsuite'] = 'autopkgtest8'
            updater.changes()
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest8
Uploaders: @lintian-brush-test@
""", "debian/control.in", strip_trailing_whitespace=True)
        self.assertFalse(os.path.exists('debian/control'))

    def test_update_template_invalid_tokens(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
# DO NOT EDIT
# This file was generated by blah

Source: blah
Testsuite: autopkgtest
Uploaders: Jelmer Vernooij <jelmer@jelmer.uk>
"""), ('debian/control.in', """\
Source: blah
Testsuite: autopkgtest
@OTHERSTUFF@
"""), ('debian/rules', """\
#!/usr/bin/make -f

debian/control: debian/control.in
\tsed -e 's/@OTHERSTUFF@/Vcs-Git: example.com/' < $< > $@
""")])
        os.chmod('debian/rules', 0o755)

        with ControlEditor() as updater:
            updater.source['Testsuite'] = 'autopkgtest8'
            updater.changes()
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest8
@OTHERSTUFF@
""", "debian/control.in", strip_trailing_whitespace=True)
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest8
Vcs-Git: example.com
""", "debian/control", strip_trailing_whitespace=True)

    def test_update_cdbs_template(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest
Build-Depends: some-foo, libc6

"""), ('debian/control.in', """\
Source: blah
Testsuite: autopkgtest
Build-Depends: @cdbs@, libc6

""")])

        with ControlEditor() as updater:
            updater.source['Build-Depends'] = 'some-foo, libc6, some-bar'
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
Build-Depends: some-foo, libc6, some-bar
""", "debian/control", strip_trailing_whitespace=True)
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
Build-Depends: @cdbs@, libc6, some-bar
""", "debian/control.in", strip_trailing_whitespace=True)

    def test_description_stays_last(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Testsuite: autopkgtest

Package: libblah
Section: extra
Description: foo
 bar

""")])

        def add_header(control):
            control["Arch"] = "all"
        self.assertTrue(update_control(binary_package_cb=add_header))
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest

Package: libblah
Section: extra
Arch: all
Description: foo
 bar
""", 'debian/control', strip_trailing_whitespace=True)

    def test_no_new_heading_whitespace(self):
        self.build_tree_contents([('debian/', ), ('debian/control', """\
Source: blah
Build-Depends:
 debhelper-compat (= 11),
 uuid-dev
""")])

        with ControlEditor() as updater:
            updater.source['Build-Depends'] = (
                '\n debhelper-compat (= 12),\n uuid-dev')
        self.assertFileEqual("""\
Source: blah
Build-Depends:
 debhelper-compat (= 12),
 uuid-dev
""", 'debian/control', strip_trailing_whitespace=True)


class ParseRelationsTests(TestCase):

    def test_empty(self):
        self.assertEqual([], parse_relations(''))
        self.assertEqual([('\n', [], '')], parse_relations('\n'))

    def test_simple(self):
        self.assertEqual(
                [('', [PkgRelation('debhelper')], '')],
                parse_relations('debhelper'))
        self.assertEqual(
                [('  \n', [PkgRelation('debhelper')], '')],
                parse_relations('  \ndebhelper'))
        self.assertEqual(
                [('  \n', [PkgRelation('debhelper')], ' \n')],
                parse_relations('  \ndebhelper \n'))


class FormatRelationsTests(TestCase):

    def test_empty(self):
        self.assertEqual(
                '',
                format_relations([('', [], '')]))
        self.assertEqual(
                '',
                format_relations([('', [], '\n')]))
        self.assertEqual(
                '',
                format_relations([('\n ', [], '')]))

    def test_simple(self):
        self.assertEqual(
                'debhelper',
                format_relations([('', [PkgRelation('debhelper')], '')]))
        self.assertEqual(
                format_relations([('  \n', [PkgRelation('debhelper')], '')]),
                '  \ndebhelper')
        self.assertEqual(
                format_relations(
                    [('  \n', [PkgRelation('debhelper')], ' \n')]),
                '  \ndebhelper ')

    def test_multiple(self):
        self.assertEqual(
                'debhelper, blah',
                format_relations([('', [PkgRelation('debhelper')], ''),
                                 (' ', [PkgRelation('blah')], '')]))


class EnsureMinimumVersionTests(TestCase):

    def test_added(self):
        self.assertEqual(
            'debhelper (>= 9)', ensure_minimum_version('', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_minimum_version('blah', 'debhelper', '9'))

    def test_unchanged(self):
        self.assertEqual(
            'debhelper (>= 9)', ensure_minimum_version(
                'debhelper (>= 9)', 'debhelper', '9'))
        self.assertEqual(
            'debhelper (= 9)', ensure_minimum_version(
                'debhelper (= 9)', 'debhelper', '9'))
        self.assertEqual(
            'debhelper (>= 9)', ensure_minimum_version(
                'debhelper (>= 9)', 'debhelper', '9~'))

    def test_updated(self):
        self.assertEqual(
            'debhelper (>= 9)',
            ensure_minimum_version('debhelper', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_minimum_version('blah, debhelper', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_minimum_version('blah, debhelper (>= 8)', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_minimum_version(
                'blah, debhelper (>= 8), debhelper (>= 8.1) | dh-systemd',
                'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (>= 9), debhelper (>= 10) | dh-systemd',
            ensure_minimum_version(
                'blah, debhelper (>= 8), debhelper (>= 10) | dh-systemd',
                'debhelper', '9'))


class EnsureRelationTests(TestCase):

    def test_added(self):
        self.assertEqual(
            'debhelper (>= 9)', ensure_relation('', 'debhelper (>= 9)'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_relation('blah', 'debhelper (>= 9)'))

    def test_unchanged(self):
        self.assertEqual(
            'debhelper (>= 9)', ensure_relation(
                'debhelper (>= 9)', 'debhelper (>= 9)'))
        self.assertEqual(
            'debhelper (= 9)', ensure_relation(
                'debhelper (= 9)', 'debhelper (>= 9)'))
        self.assertEqual(
            'debhelper (>= 9)', ensure_relation(
                'debhelper (>= 9)', 'debhelper (>= 9~)'))

    def test_updated(self):
        self.assertEqual(
            'debhelper (>= 9)',
            ensure_relation('debhelper', 'debhelper (>= 9)'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_relation('blah, debhelper', 'debhelper (>= 9)'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_relation('blah, debhelper (>= 8)', 'debhelper (>= 9)'))
        self.assertEqual(
            'blah, debhelper (>= 9)',
            ensure_relation(
                'blah, debhelper (>= 8), debhelper (>= 8.1) | dh-systemd',
                'debhelper (>= 9)'))
        self.assertEqual(
            'blah, debhelper (>= 9), debhelper (>= 10) | dh-systemd',
            ensure_relation(
                'blah, debhelper (>= 8), debhelper (>= 10) | dh-systemd',
                'debhelper (>= 9)'))


class EnsureSomeVersionTests(TestCase):

    def test_added(self):
        self.assertEqual(
            'debhelper', ensure_some_version('', 'debhelper'))
        self.assertEqual(
            'blah, debhelper',
            ensure_some_version('blah', 'debhelper'))

    def test_unchanged(self):
        self.assertEqual(
            'debhelper (>= 9)', ensure_some_version(
                'debhelper (>= 9)', 'debhelper'))
        self.assertEqual(
            'debhelper (= 9)', ensure_some_version(
                'debhelper (= 9)', 'debhelper'))
        self.assertEqual(
            'debhelper (>= 9)', ensure_some_version(
                'debhelper (>= 9)', 'debhelper'))
        self.assertEqual(
            'debhelper', ensure_some_version(
                'debhelper', 'debhelper'))


class EnsureExactVersionTests(TestCase):

    def test_added(self):
        self.assertEqual(
            'debhelper (= 9)', ensure_exact_version('', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (= 9)',
            ensure_exact_version('blah', 'debhelper', '9'))

    def test_unchanged(self):
        self.assertEqual(
            'debhelper (= 9)', ensure_exact_version(
                'debhelper (= 9)', 'debhelper', '9'))

    def test_updated(self):
        self.assertEqual(
            'debhelper (= 9)', ensure_exact_version(
                'debhelper (>= 9)', 'debhelper', '9'))
        self.assertEqual(
            'debhelper (= 9)',
            ensure_exact_version('debhelper', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (= 9)',
            ensure_exact_version('blah, debhelper', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (= 9)',
            ensure_exact_version('blah, debhelper (= 8)', 'debhelper', '9'))
        self.assertEqual(
            'blah, debhelper (= 9)',
            ensure_exact_version('blah, debhelper (= 10)', 'debhelper', '9'))
        self.assertEqual("""
 debhelper-compat (= 12),
 pkg-config,
 uuid-dev""", ensure_exact_version("""
 debhelper-compat (= 11),
 pkg-config,
 uuid-dev""", 'debhelper-compat', '12'))


class DropDependencyTests(TestCase):

    def test_deleted(self):
        self.assertEqual(
            'debhelper (>= 9)',
            drop_dependency('debhelper (>= 9), dh-autoreconf',
                            'dh-autoreconf'))
        self.assertEqual(
            'dh-autoreconf',
            drop_dependency('debhelper (>= 9), dh-autoreconf', 'debhelper'))
        self.assertEqual(
            '', drop_dependency('debhelper (>= 9)', 'debhelper'))
        self.assertEqual(
            'debhelper-compat (= 12)',
            drop_dependency(
                'debhelper (>= 9), debhelper-compat (= 12)',
                'debhelper'))


class AddDependencyTests(TestCase):

    def test_added(self):
        self.assertEqual(
            'debhelper (>= 9), dh-autoreconf',
            add_dependency('debhelper (>= 9)', 'dh-autoreconf'))
        self.assertEqual(
            'debhelper (>= 9), ${misc:Depends}',
            add_dependency('debhelper (>= 9)', '${misc:Depends}'))
        self.assertEqual(
            'debhelper (>= 9), blah,',
            add_dependency('debhelper (>= 9),', 'blah'))

    def test_indentation(self):
        self.assertEqual("""foo,
    bar,
    blah""", add_dependency("""foo,
    bar""", 'blah'))
        self.assertEqual("""foo,
 bar,
 blah""", add_dependency("""foo,
 bar""", 'blah'))
        self.assertEqual("""foo,
 bar,
 blah""", add_dependency("""foo,
 bar
""", 'blah'))

    def test_insert(self):
        self.assertEqual("""blah,
    foo,
    bar""", add_dependency("""foo,
    bar""", 'blah', position=0))
        self.assertEqual("""foo,
    blah,
    bar""", add_dependency("""foo,
    bar""", 'blah', position=1))

    def test_odd_syntax(self):
        self.assertEqual("""
 foo
 , bar
 , blah""", add_dependency("""
 foo
 , bar
""", 'blah'))
        self.assertEqual("""
 foo
 , blah
 , bar""", add_dependency("""
 foo
 , bar
""", 'blah', position=1))


class GetRelationTests(TestCase):

    def test_missing(self):
        self.assertRaises(
            KeyError, get_relation, '', 'debhelper')
        self.assertRaises(
            KeyError,
            get_relation, 'blah', 'debhelper')

    def test_simple(self):
        self.assertEqual(
            (0, [PkgRelation('debhelper', ('>=', '9'))]),
            get_relation(
                'debhelper (>= 9)', 'debhelper'))
        self.assertEqual(
            (1, [PkgRelation('debhelper', ('=', '9'))]),
            get_relation(
                'blah, debhelper (= 9)', 'debhelper'))

    def test_complex(self):
        self.assertRaises(
            ValueError,
            get_relation, 'blah | debhelper (= 9)', 'debhelper')
        self.assertRaises(
            ValueError,
            get_relation,
            'blah, debhelper (= 9) | debhelper (<< 10)', 'debhelper')


class IterRelationsTests(TestCase):

    def test_missing(self):
        self.assertEqual(
            [], list(iter_relations('', 'debhelper')))
        self.assertEqual(
            [], list(iter_relations('blah', 'debhelper')))

    def test_simple(self):
        self.assertEqual(
            [(0, [PkgRelation('debhelper', ('>=', '9'))])],
            list(iter_relations(
                'debhelper (>= 9)', 'debhelper')))
        self.assertEqual(
            [(1, [PkgRelation('debhelper', ('=', '9'))])],
            list(iter_relations(
                'blah, debhelper (= 9)', 'debhelper')))

    def test_complex(self):
        self.assertEqual(
            [(0, [PkgRelation('blah'), PkgRelation('debhelper', ('=', '9'))])],
            list(iter_relations('blah | debhelper (= 9)', 'debhelper')))
        self.assertEqual(
            [(1,
              [PkgRelation('debhelper', ('=', '9')),
               PkgRelation('debhelper', ('<<', '10'))])],
            list(iter_relations(
                'blah, debhelper (= 9) | debhelper (<< 10)', 'debhelper')))


class DeleteFromListTests(TestCase):

    def test_intermediate(self):
        self.assertEqual('a, c', delete_from_list('a, b, c', 'b'))
        self.assertEqual('a, c', delete_from_list('a, b, c', 'b '))

    def test_head(self):
        self.assertEqual('b, c', delete_from_list('a, b, c', 'a'))
        self.assertEqual(' b, c', delete_from_list(' a, b, c', 'a'))

    def test_tail(self):
        self.assertEqual('a, b', delete_from_list('a, b, c', 'c'))
        self.assertEqual('a, b', delete_from_list('a, b , c', 'c'))

    def test_only(self):
        self.assertEqual('a', delete_from_list('a', 'c'))
        self.assertEqual('', delete_from_list('a', 'a'))


class IsDepImpliedTests(TestCase):

    def parse(self, p):
        [dep] = PkgRelation.parse(p)
        return dep

    def test_no_version(self):
        self.assertTrue(
            is_dep_implied(self.parse('bzr'), self.parse('bzr')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr'), self.parse('bzr (>= 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr'), self.parse('bzr (<< 3)')))

    def test_wrong_package(self):
        self.assertFalse(
            is_dep_implied(self.parse('bzr'), self.parse('foo (<< 3)')))

    def test_version(self):
        self.assertFalse(
            is_dep_implied(self.parse('bzr (>= 3)'), self.parse('bzr (<< 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (>= 3)'), self.parse('bzr (= 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (= 3)'), self.parse('bzr (>= 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (>= 3)'), self.parse('bzr (>> 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (= 3)'), self.parse('bzr (= 4)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (>= 3)'), self.parse('bzr (>= 2)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (>= 3)'), self.parse('bzr (>= 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr'), self.parse('bzr (<< 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (<< 3)'), self.parse('bzr (<< 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (<= 3)'), self.parse('bzr (<< 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (>= 2)'), self.parse('bzr (<< 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (<< 2)'), self.parse('bzr (<< 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (<= 2)'), self.parse('bzr (<< 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (<= 5)'), self.parse('bzr (<< 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (<= 5)'), self.parse('bzr (= 3)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (<= 5)'), self.parse('bzr (>= 3)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (>> 5)'), self.parse('bzr (>> 6)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (>> 5)'), self.parse('bzr (>> 5)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (>> 5)'), self.parse('bzr (>> 4)')))
        self.assertTrue(
            is_dep_implied(self.parse('bzr (>> 5)'), self.parse('bzr (= 6)')))
        self.assertFalse(
            is_dep_implied(self.parse('bzr (>> 5)'), self.parse('bzr (= 5)')))


class IsRelationImpliedTests(TestCase):

    def test_unrelated(self):
        self.assertFalse(is_relation_implied('bzr', 'bar'))
        self.assertFalse(is_relation_implied('bzr (= 3)', 'bar'))
        self.assertFalse(is_relation_implied('bzr (= 3) | foo', 'bar'))

    def test_too_old(self):
        self.assertFalse(is_relation_implied('bzr (= 3)', 'bzr'))
        self.assertFalse(is_relation_implied('bzr (= 3)', 'bzr (= 2)'))
        self.assertFalse(is_relation_implied('bzr (= 3)', 'bzr (>= 2)'))

    def test_ors(self):
        self.assertFalse(is_relation_implied('bzr (= 3)', 'bzr | foo'))
        self.assertFalse(is_relation_implied('bzr', 'bzr | foo'))
        self.assertTrue(is_relation_implied('bzr | foo', 'bzr | foo'))

    def test_implied(self):
        self.assertTrue(is_relation_implied('bzr (= 3)', 'bzr (= 3)'))
        self.assertTrue(is_relation_implied('bzr (>= 3)', 'bzr (>= 4)'))
        self.assertTrue(is_relation_implied('bzr (>= 4)', 'bzr (>= 4)'))
        self.assertTrue(is_relation_implied('bzr', 'bzr'))
        self.assertTrue(is_relation_implied('bzr | foo', 'bzr'))
        self.assertFalse(is_relation_implied('bzr (= 3)', 'bzr (>= 3)'))


class CdbsResolverConflictTests(TestCase):

    def test_build_depends(self):
        val = _cdbs_resolve_conflict(
            ('Source', 'libnetsds-perl'), 'Build-Depends',
            'debhelper (>= 6), foo', '@cdbs@', 'debhelper (>= 10), foo')
        self.assertEqual(val, '@cdbs@, debhelper (>= 10)')
        val = _cdbs_resolve_conflict(
            ('Source', 'libnetsds-perl'), 'Build-Depends',
            'debhelper (>= 6), foo',  '@cdbs@, foo',
            'debhelper (>= 10), foo')
        self.assertEqual(val, '@cdbs@, foo, debhelper (>= 10)')
        val = _cdbs_resolve_conflict(
            ('Source', 'libnetsds-perl'), 'Build-Depends',
            'debhelper (>= 6), foo', '@cdbs@, debhelper (>= 9)',
            'debhelper (>= 10), foo')
        self.assertEqual(val, '@cdbs@, debhelper (>= 10)')


class ParseStandardsVersionTests(TestCase):

    def test_parse(self):
        self.assertEqual((4, 5, 0), parse_standards_version('4.5.0'))


class GuessTemplateTypeTests(TestCaseInTempDir):

    def setUp(self):
        super().setUp()
        os.mkdir('debian')

    def test_rules_generates_control(self):
        with open('debian/rules', 'w') as f:
            f.write("""\
%:
    dh $@

debian/control: debian/control.in
    cp $@ $<
""")
        self.assertEqual(
            'rules', guess_template_type('debian/control.in', 'debian'))

        with open('debian/rules', 'w') as f:
            f.write("""\
%:
    dh $@

debian/%: debian/%.in
    cp $@ $<
""")
        self.assertEqual(
            'rules', guess_template_type('debian/control.in', 'debian'))

    def test_blends(self):
        with open('debian/rules', 'w') as f:
            f.write("""\
%:
    dh $@

include /usr/share/blends-dev/rules
""")
        self.assertEqual(
            'rules', guess_template_type('debian/control.stub', 'debian'))

    def test_no_build_depends(self):
        with open('debian/control.in', 'w') as f:
            # No paragraph
            f.write("")

        self.assertIs(None, guess_template_type('debian/control.in', 'debian'))

        with open('debian/control.in', 'w') as f:
            f.write("""\
Source: blah
Build-Depends: cdbs
Vcs-Git: file://

Package: bar
""")
        self.assertEqual(
            "cdbs",
            guess_template_type('debian/control.in', 'debian'))

        with open('debian/control.in', 'w') as f:
            f.write("""\
Source: blah
Vcs-Git: file://

Package: bar
""")
        self.assertIs(None, guess_template_type('debian/control.in', 'debian'))

    def test_gnome(self):
        with open('debian/control.in', 'w') as f:
            f.write("""\
Foo @GNOME_TEAM@
""")
        self.assertEqual(
            'gnome', guess_template_type('debian/control.in'))

    def test_gnome_build_depends(self):
        with open('debian/control.in', 'w') as f:
            f.write("""\
Source: blah
Build-Depends: gnome-pkg-tools, libc6-dev
""")
        self.assertEqual(
            'gnome', guess_template_type('debian/control.in'))

    def test_cdbs(self):
        with open('debian/control.in', 'w') as f:
            f.write("""\
Source: blah
Build-Depends: debhelper, cdbs
""")
        self.assertEqual(
            'cdbs', guess_template_type('debian/control.in'))

    def test_multiple_paragraphs(self):
        with open('debian/control.in', 'w') as f:
            f.write("""\
Source: blah
Build-Depends: debhelper, cdbs

Package: foo
""")
        self.assertEqual(
            'cdbs', guess_template_type('debian/control.in'))

    def test_directory(self):
        os.mkdir('debian/control.in')
        self.assertEqual('directory', guess_template_type('debian/control.in'))

    def test_debcargo(self):
        with open('debian/control.in', 'w') as f:
            f.write("""\
Source: blah
Build-Depends: bar
""")
        with open('debian/debcargo.toml', 'w') as f:
            f.write('maintainer = Joe Example <joe@example.com>\n')
        self.assertEqual(
            'debcargo', guess_template_type('debian/control.in', 'debian'))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.debcargo."""

from contextlib import ExitStack
import os
import shutil
import tempfile
from unittest import TestCase


from debmutate.debcargo import (
    debcargo_version_to_semver,
    semver_pair,
    DebcargoControlShimEditor,
    DebcargoEditor,
    DEFAULT_MAINTAINER,
)


class DebcargoVersionToSemverTests(TestCase):

    def test_prerelease(self):
        self.assertEqual(
            '1.0.0-rc1', debcargo_version_to_semver('1.0.0~rc1'))


class SemverPairTests(TestCase):

    def test_pair(self):
        self.assertEqual('1.2', semver_pair('1.2.3-rc1'))
        self.assertEqual('1.2', semver_pair('1.2.3'))
        self.assertEqual('1.2', semver_pair('1.2.4+ds'))


class DebcargoControlShimEditorTests(TestCase):

    def setUp(self):
        super().setUp()
        self.td = tempfile.mkdtemp()
        os.mkdir(os.path.join(self.td, 'debian'))
        self.addCleanup(os.chdir, os.getcwd())
        os.chdir(self.td)
        self.addCleanup(shutil.rmtree, self.td)
        self.debcargo = DebcargoEditor(allow_missing=True)
        self.es = ExitStack()
        self.es.enter_context(self.debcargo)
        self.editor = DebcargoControlShimEditor(
            self.debcargo, 'crate', '1.2.3',
            features=['feature1', 'feature2'])
        self.es.enter_context(self.editor)

    def test_properties(self):
        self.assertEqual(['feature1', 'feature2'], self.editor.features)
        self.assertEqual('crate', self.editor.crate_name)
        self.assertEqual('1.2.3', self.editor.crate_version)

    def test_source(self):
        self.assertEqual('rust-crate', self.editor.source['Source'])

    def test_source_semver_suffix(self):
        self.debcargo['semver_suffix'] = True
        self.assertEqual('rust-crate-1.2', self.editor.source['Source'])

    def test_source_priority(self):
        self.assertEqual('optional', self.editor.source['Priority'])

    def test_source_vcs_git(self):
        self.assertEqual(
            'https://salsa.debian.org/rust-team/debcargo-conf.git [src/crate]',
            self.editor.source['Vcs-Git'])

    def test_source_vcs_git_specified(self):
        self.debcargo['source'] = {
            'vcs_git': 'https://github.com/jelmer/crate'}
        self.assertEqual(
            'https://github.com/jelmer/crate', self.editor.source['Vcs-Git'])

    def test_source_vcs_browser(self):
        self.assertEqual(
            'https://salsa.debian.org/rust-team/'
            'debcargo-conf/tree/master/src/crate',
            self.editor.source['Vcs-Browser'])

    def test_source_vcs_browser_specified(self):
        self.debcargo['source'] = {
            'vcs_browser': 'https://github.com/jelmer/crate'}
        self.assertEqual(
            'https://github.com/jelmer/crate',
            self.editor.source['Vcs-Browser'])

    def test_homepage(self):
        self.assertRaises(KeyError, self.editor.source.__getitem__, 'Homepage')

    def test_homepage_specified(self):
        self.debcargo['source'] = {
            'homepage': 'https://github.com/jelmer/crate'}
        self.assertEqual(
            'https://github.com/jelmer/crate', self.editor.source['Homepage'])

    def test_section(self):
        self.assertEqual('rust', self.editor.source['Section'])

    def test_section_specified(self):
        self.debcargo['source'] = {'section': 'web'}
        self.assertEqual('web', self.editor.source['Section'])

    def test_standards_version(self):
        self.assertIsInstance(self.editor.source['Standards-Version'], str)

    def test_standards_version_specified(self):
        self.debcargo['source'] = {'policy': '1.2.3'}
        self.assertEqual('1.2.3', self.editor.source['Standards-Version'])

    def test_maintainer(self):
        self.assertEqual(DEFAULT_MAINTAINER, self.editor.source['Maintainer'])

    def test_maintainer_specified(self):
        self.debcargo['maintainer'] = 'Jelmer Vernooij <jelmer@example.com>'
        self.assertEqual(
            'Jelmer Vernooij <jelmer@example.com>',
            self.editor.source['Maintainer'])

    def test_uploaders(self):
        self.assertRaises(
            KeyError, self.editor.source.__getitem__, 'Uploaders')

    def test_uploaders_specified(self):
        self.debcargo['uploaders'] = 'Jelmer Vernooij <jelmer@example.com>'
        self.assertEqual(
            'Jelmer Vernooij <jelmer@example.com>',
            self.editor.source['Uploaders'])

    def test_rules_requires_root(self):
        self.assertEqual('no', self.editor.source['Rules-Requires-Root'])



#!/usr/bin/python
# Copyright (C) 2019-2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.debhelper."""

from typing import Dict

from debian.changelog import Version

from . import (
    TestCase,
    TestCaseInTempDir,
    )

from debmutate.debhelper import (
    ensure_minimum_debhelper_version,
    MaintscriptEditor,
    MaintscriptMoveConffile,
    )


class EnsureMinumumDebhelperVersionTests(TestCase):

    def test_already(self):
        d = {
            'Build-Depends': 'debhelper (>= 10)',
            }
        self.assertFalse(ensure_minimum_debhelper_version(d, '10'))
        self.assertEqual(d, {'Build-Depends': 'debhelper (>= 10)'})
        self.assertFalse(ensure_minimum_debhelper_version(d, '9'))
        self.assertEqual(d, {'Build-Depends': 'debhelper (>= 10)'})

    def test_already_compat(self):
        d = {
            'Build-Depends': 'debhelper-compat (= 10)',
            }
        self.assertFalse(ensure_minimum_debhelper_version(d, '10'))
        self.assertEqual(d, {'Build-Depends': 'debhelper-compat (= 10)'})
        self.assertFalse(ensure_minimum_debhelper_version(d, '9'))
        self.assertEqual(d, {'Build-Depends': 'debhelper-compat (= 10)'})

    def test_bump(self):
        d = {
            'Build-Depends': 'debhelper (>= 10)',
            }
        self.assertTrue(ensure_minimum_debhelper_version(d, '11'))
        self.assertEqual(d, {'Build-Depends': 'debhelper (>= 11)'})

    def test_bump_compat(self):
        d = {
            'Build-Depends': 'debhelper-compat (= 10)',
            }
        self.assertTrue(ensure_minimum_debhelper_version(d, '11'))
        self.assertEqual(d, {
            'Build-Depends': 'debhelper-compat (= 10), debhelper (>= 11)'})
        self.assertTrue(ensure_minimum_debhelper_version(d, '11.1'))
        self.assertEqual(d, {
            'Build-Depends': 'debhelper-compat (= 10), debhelper (>= 11.1)'})

    def test_not_set(self):
        d: Dict[str, str] = {}
        self.assertTrue(ensure_minimum_debhelper_version(d, '10'))
        self.assertEqual(d, {'Build-Depends': 'debhelper (>= 10)'})

    def test_in_indep(self):
        d = {'Build-Depends-Indep': 'debhelper (>= 9)'}
        self.assertRaises(Exception, ensure_minimum_debhelper_version, d, '10')


class MaintscriptEditorTests(TestCaseInTempDir):

    def test_simple_edit(self):
        self.build_tree_contents([('debian/', ), ('debian/maintscript', """\
mv_conffile /etc/iptotal/apache.conf /etc/apache2/conf-available/iptotal.conf \
0.3.3-13.1~
""")])
        with MaintscriptEditor() as e:
            self.assertEqual([MaintscriptMoveConffile(
                '/etc/iptotal/apache.conf',
                '/etc/apache2/conf-available/iptotal.conf',
                Version('0.3.3-13.1~'))], e.entries)

    def test_simple_missing(self):
        with MaintscriptEditor() as e:
            self.assertEqual([], e.entries)

    def test_simple_comment(self):
        self.build_tree_contents([('debian/', ), ('debian/maintscript', """\
# I am a comment
mv_conffile /etc/iptotal/apache.conf /etc/apache2/conf-available/iptotal.conf \
0.3.3-13.1~
""")])
        with MaintscriptEditor() as e:
            del e[0]
            self.assertRaises(IndexError, e.__delitem__, 0)
        with open('debian/maintscript') as f:
            self.assertEqual("# I am a comment\n", f.read())



#!/usr/bin/python3

# Copyright (C) 2021 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Ben file parsing."""

import re


SUPPORTED_KEYS = [
    'title', 'notes', 'is_affected', 'is_good', 'is_bad', 'export']


def _parse_benitem(v):
    if v == 'false':
        return False
    elif v == 'true':
        return True
    elif v.startswith('[') and v.endswith(']'):
        return [_parse_benitem(i) for i in v.split(';')]
    elif v.startswith('"') and v.endswith('"'):
        return v[1:-1]
    elif v.isdigit():
        return int(v)
    elif '~' in v:
        regex = []
        for o in v.split(' | '):
            try:
                field, expr = o.split('~', 1)
            except ValueError:
                raise ValueError('expected ~: %r' % o)
            expr = re.compile(_parse_benitem(expr.strip()))
            regex.append((field.strip(), expr))
        return regex
    else:
        return v


def parse_ben(f):
    ret = {}
    assignment = {}
    lastk = None
    v = ''
    for line in f:
        if lastk is None:
            if not line.strip():
                continue
            if line.startswith('#'):
                continue
            line = line.rstrip()
            (k, v) = line.split('=', 1)
            k = k.strip()
            v = v.lstrip()
            lastk = k
        else:
            v += line
        if v.rstrip().endswith(';'):
            assignment[lastk] = v[:-1]
            lastk = None
    if lastk is not None:
        raise ValueError('unterminated key: %r' % lastk)
    for k, v in assignment.items():
        ret[k] = _parse_benitem(v)
    return ret



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with Debian versions."""

from datetime import datetime
import re
from typing import Optional, Tuple, Union
from debian.changelog import Version

__all__ = [
    'git_snapshot_data_from_version',
    'mangle_version_for_git',
    'upstream_version_add_revision',
]


# Ideally we wouldn't have a list like this, but unfortunately we do.
COMMON_VENDORS = ['debian', 'ubuntu', 'kali']


DFSG_REGEX = re.compile(r'^(.*)([\+~])(dfsg|ds)([0-9]*)$')


def git_snapshot_data_from_version(
        version: Union[str, Version]) -> Tuple[Optional[str], Optional[str]]:
    """Extract git snapshot information from an upstream version string.

    Args:
      version: A version string
    """
    version = str(version)
    git_id = None
    date = None
    if "+git" in version or "~git" in version or "-git" in version:
        m = re.match(
            ".*[~+-]git([0-9]{4})([0-9]{2})([0-9]{2})\\.([0-9a-f]{7}).*",
            version)
        if not m:
            m = re.match(
                ".*[~+-]git([0-9]{4})([0-9]{2})([0-9]{2})\\."
                "[0-9+]\\.([0-9a-f]{7}).*",
                version)
        if m:
            git_id = m.group(4)
            date = "{}-{}-{}".format(m.group(1), m.group(2), m.group(3))
        else:
            m = re.match(".*[~+]git([0-9]{4})([0-9]{2})([0-9]{2}).*", version)
            if m:
                date = "{}-{}-{}".format(m.group(1), m.group(2), m.group(3))
    m = re.match(r".*\+next\.([0-9]{4})([0-9]{2})([0-9]{2}).*", version)
    if m:
        date = "{}-{}-{}".format(m.group(1), m.group(2), m.group(3))
    return (git_id, date)


def mangle_version_for_git(version: Union[Version, str]) -> str:
    version = str(version)
    # See https://dep-team.pages.debian.net/deps/dep14/
    manipulated = (
        version.replace("~", "_").replace(':', '%').replace('..', '.#.'))
    if manipulated.endswith('.'):
        manipulated += '#'
    if manipulated.endswith('.lock'):
        manipulated = manipulated[:-4] + '#lock'
    return manipulated


def initial_debian_revision(distribution_name):
    if distribution_name == "ubuntu":
        return "0ubuntu1"
    else:
        return "1"


def new_upstream_package_version(
        upstream_version, debian_revision, epoch=None):
    ret = Version("{}-{}".format(upstream_version, debian_revision))
    ret.epoch = epoch
    return ret


def new_package_version(upstream_version, distribution_name, epoch=None):
    """Determine the package version for a new upstream.

    Args:
      upstream_version: Upstream version string
      distribution_name: Distribution the package is for
      epoch: Optional epoch
    """
    debian_revision = initial_debian_revision(distribution_name)
    return new_upstream_package_version(
        upstream_version, debian_revision, epoch=epoch)


def get_snapshot_revision(upstream_version):
    """Return the upstream revision specifier if specified in the upstream
    version.

    When packaging an upstream snapshot some people use +vcsnn or ~vcsnn to
    indicate what revision number of the upstream VCS was taken for the
    snapshot. This given an upstream version number this function will return
    an identifier of the upstream revision if it appears to be a snapshot. The
    identifier is a string containing a bzr revision spec, so it can be
    transformed in to a revision.

    Args:
      upstream_version: a string containing the upstream version number.
    Returns:
      a string containing a revision specifier for the revision of the
      upstream branch that the snapshot was taken from, or None if it
      doesn't appear to be a snapshot.
    """
    match = re.search("(?:~|\\+)bzr([0-9]+)$", upstream_version)
    if match is not None:
        return ("bzr", match.groups()[0])
    match = re.search("(?:~|\\+)svn([0-9]+)$", upstream_version)
    if match is not None:
        return ("svn", match.groups()[0])
    match = re.match(r"^(.*)([\+~])git(\d{8})\.([a-f0-9]{7})$",
                     upstream_version)
    if match:
        return ("git", match.group(4))
    match = re.match(r"^(.*)([\+~])git(\d{8})$", upstream_version)
    if match:
        return ("date", match.group(3))
    return None


def upstream_version_add_revision(
        version_string: str, sep: str = '+',
        gitid: Optional[bytes] = None,
        gitdate: Optional[datetime] = None,
        bzr_revno: Optional[str] = None,
        svn_revno: Optional[int] = None):
    """Update the revision in a upstream version string.

    Args:
      branch: Branch in which the revision can be found
      version_string: Original version string
      sep: Separator to use when adding snapshot
      gitid: git sha (as bytes)
      gitdate: timestamp for git
      bzr_revno: Bazaar dotted revno
      svn_revno: Subversion revision number
    """
    m = DFSG_REGEX.match(version_string)
    if m:
        version_string = m.group(1)
    if bzr_revno is not None:
        m = re.match(r"^(.*)([\+~])bzr(\d+)$", version_string)
        if m:
            return "{}{}bzr{}".format(m.group(1), m.group(2), bzr_revno)

    if gitid:
        decoded_gitid: Optional[str] = gitid[:7].decode('ascii')
    else:
        decoded_gitid = None
    if gitdate:
        gitdate_formatted: Optional[str] = gitdate.strftime('%Y%m%d')
    else:
        gitdate_formatted = None

    m = re.match(r"^(.*)([\+~-])git(\d{8})\.([a-f0-9]{7})$", version_string)
    if m and decoded_gitid:
        return "{}{}git{}.{}".format(
            m.group(1), m.group(2), gitdate_formatted, decoded_gitid)

    m = re.match(r"^(.*)([\+~-])git(\d{8})\.(\d+)\.([a-f0-9]{7})$",
                 version_string)
    if m and decoded_gitid:
        if gitdate_formatted == m.group(3):
            snapshot = int(m.group(4)) + 1
        else:
            snapshot = 0
        return "%s%sgit%s.%d.%s" % (
            m.group(1), m.group(2), gitdate_formatted, snapshot, decoded_gitid)

    m = re.match(r"^(.*)([\+~-])git(\d{8})$", version_string)
    if m and decoded_gitid:
        return "{}{}git{}".format(m.group(1), m.group(2), gitdate_formatted)

    m = re.match(r"^(.*)([\+~])svn(\d+)$", version_string)
    # FIXME: Raise error if +svn/~svn is present and svn_revno is not set?
    if m and svn_revno:
        return "%s%ssvn%d" % (m.group(1), m.group(2), svn_revno)

    if svn_revno:
        return "%s%ssvn%d" % (version_string, sep, svn_revno)
    elif decoded_gitid:
        return "{}{}git{}.1.{}".format(
            version_string, sep, gitdate_formatted, decoded_gitid)
    elif bzr_revno is not None:
        return "{}{}bzr{}".format(version_string, sep, bzr_revno)
    else:
        raise ValueError


def debianize_upstream_version(
        version: str, package: Optional[str] = None) -> str:
    """Make an upstream version string suitable for Debian.

    Args:
      version: original upstream version string
      package: package name, if known
    Returns:
      mangled version string for use in Debian versions
    """
    if version.count('_') == 1 and version.count('.') > 0:
        # This is a style commonly used for perl packages.
        # Most debian packages seem to just drop the underscore.
        # See
        # http://blogs.perl.org/users/grinnz/2018/04/a-guide-to-versions-in-perl.html
        version = version.replace('_', '')
    if '_' in version and '.' not in version:
        version = version.replace('_', '.')
    version = version.replace('-rc', '~rc')
    version = version.replace('-beta', '~beta')
    version = version.replace('-alpha', '~alpha')
    m = re.fullmatch(r'(.*)\.([0-9])(a|b|rc|alpha|beta)([0-9]*)', version)
    if m:
        version = m.group(1) + '.' + m.group(2) + '~' + m.group(3) + m.group(4)
    return version


def matches_release(upstream_version: str, release_version: str) -> bool:
    """Check whether an upstream version string matches a upstream release.

    This will e.g. strip git and dfsg suffixes before comparing.

    Args:
      upstream_version: Upstream version string
      release_version: Release to check for
    """
    release_version = release_version.lower()
    upstream_version = upstream_version.lower()
    if upstream_version == release_version:
        return True
    m = re.match("(.*)([~+-])(ds|dfsg|git|bzr|svn|hg).*", upstream_version)
    if m and m.group(1) == release_version:
        return True
    m = re.match("(.*)([~+-]).*", upstream_version)
    if m and m.group(1) == release_version:
        return True
    m = re.match('.*~([0-9.]+)$', upstream_version)
    if m and m.group(1) == release_version:
        return True
    return False


def strip_dfsg_suffix(version: str) -> str:
    """Strip the DFSG suffix from a version."""
    m = DFSG_REGEX.match(version)
    if m:
        return m.group(1)
    return version


def add_dfsg_suffix(
        upstream_version: str,
        old_upstream_version: Optional[str] = None) -> str:
    """Add a dfsg suffix to an version version string.

    Allow old_upstream_version to be passed in so optionally the format can be
    kept consistent.
    """
    style = "+ds"
    if old_upstream_version:
        m = DFSG_REGEX.match(old_upstream_version)
        if m:
            style = m.group(2) + m.group(3)
            if m.group(4):
                style += "1"
    return upstream_version + style



#!/usr/bin/python3
# Copyright (C) 2022 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with changelog files."""

__all__ = [
    'get_vendor_name',
]


import os
from debian.deb822 import Deb822


def _load_vendor_file(vendor: str = 'default') -> Deb822:
    with open('/etc/dpkg/origins/%s' % vendor) as f:
        return Deb822(f)


def get_vendor_name() -> str:
    if 'DEB_VENDOR' in os.environ:
        return os.environ['DEB_VENDOR']

    vendor = _load_vendor_file('default')
    return vendor['Vendor']



#!/usr/bin/python3
# PkgRelation from python-debian, lib/debian/deb822.py (GPL-2+).
# This file is a part of debmutate.
# Local changes to add support for preserving whitespace
# Copyright (C) 2005-2006  dann frazier <dannf@dannf.org>
# Copyright (C) 2006-2010  John Wright <john@johnwright.org>
# Copyright (C) 2006       Adeodato Simó <dato@net.com.org.es>
# Copyright (C) 2008       Stefano Zacchiroli <zack@upsilon.cc>
# Copyright (C) 2014       Google, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# TODO(jelmer): Contribute improvements back to python-debian

import collections
import logging
import re


class PkgRelation:
    """A package requirement."""

    __dep_RE = re.compile(
        r'^\s*(?P<name>[a-zA-Z0-9.+\-]{2,})'
        r'(:(?P<archqual>([a-zA-Z0-9][a-zA-Z0-9-]*)))?'
        r'(\s*\(\s*(?P<relop>[>=<]+)\s*'
        r'(?P<version>[0-9a-zA-Z:\-+~.]+)\s*\))?'
        r'(\s*\[(?P<archs>[\s!\w\-]+)\])?\s*'
        r'((?P<restrictions><.+>))?\s*'
        r'$')
    __pipe_sep_RE = re.compile(r'\s*\|\s*')
    __blank_sep_RE = re.compile(r'\s+')
    __restriction_sep_RE = re.compile(r'>\s*<')
    __restriction_RE = re.compile(
        r'(?P<enabled>\!)?'
        r'(?P<profile>[^\s]+)')

    ArchRestriction = collections.namedtuple('ArchRestriction',
                                             ['enabled', 'arch'])
    BuildRestriction = collections.namedtuple('BuildRestriction',
                                              ['enabled', 'profile'])

    @classmethod
    def parse(cls, text):
        def parse_archs(raw):
            # type: (str) -> list[PkgRelation.ArchRestriction]
            # assumption: no space between '!' and architecture name
            archs = []
            for arch in cls.__blank_sep_RE.split(raw.strip()):
                disabled = arch[0] == '!'
                if disabled:
                    arch = arch[1:]
                archs.append(cls.ArchRestriction(not disabled, arch))
            return archs

        def parse_restrictions(raw):
            # type: (str) -> list[list[PkgRelation.BuildRestriction]]
            """ split a restriction formula into a list of restriction lists

            Each term in the restriction list is a namedtuple of form:

                (enabled, label)

            where
                enabled: bool: whether the restriction is positive or negative
                profile: the profile name of the term e.g. 'stage1'
            """
            restrictions = []
            groups = cls.__restriction_sep_RE.split(raw.lower().strip('<> '))
            for rgrp in groups:
                group = []
                for restriction in cls.__blank_sep_RE.split(rgrp):
                    match = cls.__restriction_RE.match(restriction)
                    if match:
                        parts = match.groupdict()
                        group.append(
                            cls.BuildRestriction(
                                parts['enabled'] != '!',
                                parts['profile'],
                            ))
                restrictions.append(group)
            return restrictions

        def parse_rel(raw):
            match = cls.__dep_RE.match(raw)
            if match:
                parts = match.groupdict()
                d = {
                    'name': parts['name'],
                    'archqual': parts['archqual'],
                    'version': None,
                    'arch': None,
                    'restrictions': None,
                }
                if parts['relop'] or parts['version']:
                    d['version'] = (parts['relop'], parts['version'])
                if parts['archs']:
                    d['arch'] = parse_archs(parts['archs'])
                if parts['restrictions']:
                    d['restrictions'] = parse_restrictions(
                        parts['restrictions'])
                return PkgRelation(**d)

            logging.debug(
                'cannot parse package'
                ' relationship "%s", returning it raw' % raw)
            return PkgRelation(
                name=raw,
                version=None,
                arch=None
            )
        if text == "":
            return []
        or_deps = cls.__pipe_sep_RE.split(text)
        return [parse_rel(or_dep) for or_dep in or_deps]

    def __repr__(self):
        return "{}({!r}, {!r}, {!r}, {!r}, {!r})".format(
            self.__class__.__name__, self.name, self.version, self.arch,
            self.archqual, self.restrictions)

    def __tuple__(self):
        return (self.name, self.version, self.arch, self.archqual,
                self.restrictions)

    def __eq__(self, other):
        if not isinstance(other, PkgRelation):
            return False
        return (self.__tuple__() == other.__tuple__())

    def __lt__(self, other):
        if not isinstance(other, PkgRelation):
            raise TypeError
        return (self.__tuple__() < other.__tuple__())

    def str(self):
        """Format to string structured inter-package relationships

        Perform the inverse operation of parse_relations, returning a string
        suitable to be written in a package stanza.
        """
        def pp_arch(arch_spec):
            # type: (PkgRelation.ArchRestriction) -> str
            return '{}{}'.format(
                '' if arch_spec.enabled else '!',
                arch_spec.arch,
            )

        def pp_restrictions(restrictions):
            # type: (list[PkgRelation.BuildRestriction]) -> str
            s = []
            for term in restrictions:
                s.append(
                    '{}{}'.format(
                        '' if term.enabled else '!',
                        term.profile
                    )
                )
            return '<%s>' % ' '.join(s)

        s = self.name
        if self.archqual is not None:
            s += ':%s' % self.archqual
        if self.version is not None:
            s += ' (%s %s)' % self.version
        if self.arch is not None:
            s += ' [%s]' % ' '.join(map(pp_arch, self.arch))
        if self.restrictions is not None:
            s += ' %s' % ' '.join(map(pp_restrictions,
                                      self.restrictions))
        return s

    def __init__(self, name, version=None, arch=None, archqual=None,
                 restrictions=None):
        self.name = name
        self.version = version
        self.arch = arch
        self.archqual = archqual
        self.restrictions = restrictions



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# Tests for find_thanks and find_extra_authors originally imported from
# breezy-debian, also (C) 2006 James Westby
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.changelog."""

from datetime import datetime
from debian.changelog import Changelog
import os
import shutil
import tempfile

from debmutate.changelog import (
    ChangelogCreateError,
    ChangelogEditor,
    TextWrapper,
    all_sha_prefixed,
    rewrap_change,
    find_extra_authors,
    find_thanks,
    increment_version,
    changes_sections,
    release,
    strip_changelog_message,
    new_upstream_package_version,
    changeblock_ensure_first_line,
    find_last_distribution,
    upstream_merge_changelog_line,
    take_uploadership,
    )

from debian.changelog import Version

from unittest import TestCase


class CreateChangelogTests(TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.test_dir)
        self.addCleanup(os.chdir, os.getcwd())
        os.chdir(self.test_dir)
        os.mkdir('debian')

    def test_create(self):
        with ChangelogEditor.create() as updater:
            updater.new_block(
                package='blah', version=Version('1.0-1'),
                author='Jelmer Vernooij <jelmer@debian.org>',
                date='Mon, 02 Sep 2019 00:23:11 +0000',
                distributions='UNRELEASED',
                changes=['', '  * New entry.', ''])

        with open('debian/changelog') as f:
            self.assertEqual("""\
blah (1.0-1) UNRELEASED; urgency=unknown

  * New entry.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", f.read())

    def test_auto_version(self):
        with ChangelogEditor.create() as updater:
            updater.auto_version(
                Version('1.0-1'),
                package='blah',
                urgency='low',
                maintainer=('Jelmer Vernooij', 'jelmer@debian.org'),
                timestamp=datetime.utcfromtimestamp(1604934305))

        with open('debian/changelog') as f:
            self.assertEqual("""\
blah (1.0-1) UNRELEASED; urgency=low
 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 09 Nov 2020 15:05:05 -0000
""", f.read())


class UpdateChangelogTests(TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.test_dir)
        self.addCleanup(os.chdir, os.getcwd())
        os.chdir(self.test_dir)
        os.mkdir('debian')
        with open('debian/changelog', 'w') as f:
            f.write("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")

    def test_edit_simple(self):
        with ChangelogEditor() as updater:
            updater.changelog.version = '0.29'
        with open('debian/changelog') as f:
            self.assertEqual("""\
lintian-brush (0.29) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", f.read())

    def test_auto_version_update(self):
        with ChangelogEditor() as updater:
            updater.auto_version(
                Version('0.29'),
                timestamp=datetime.utcfromtimestamp(1604934305))
        with open('debian/changelog') as f:
            self.assertEqual("""\
lintian-brush (0.29) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 09 Nov 2020 15:05:05 -0000
""", f.read())

    def test_auto_version_add(self):
        with open('debian/changelog', 'w') as f:
            f.write("""\
lintian-brush (0.28) unstable; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")

        with ChangelogEditor() as updater:
            updater.auto_version(
                Version('0.29'),
                maintainer=('Jelmer Vernooij', 'jelmer@debian.org'),
                timestamp=datetime.utcfromtimestamp(1604934305))
        with open('debian/changelog') as f:
            self.assertEqual("""\
lintian-brush (0.29) UNRELEASED; urgency=low
 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 09 Nov 2020 15:05:05 -0000

lintian-brush (0.28) unstable; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", f.read())

    def test_invalid(self):
        with open('debian/changelog', 'w') as f:
            f.write("""\
lalalalala
""")
        self.assertRaises(ChangelogCreateError, ChangelogEditor().__enter__)

    def test_add_entry_to_new_block(self):
        with ChangelogEditor.create() as updater:
            updater[-1].distributions = 'unstable'
            updater.auto_version(
                Version('2.0-1'),
                package='blah',
                urgency='low',
                maintainer=('Jelmer Vernooij', 'jelmer@debian.org'),
                timestamp=datetime.utcfromtimestamp(1604934305))
            updater.add_entry(
                ['New release'],
                maintainer=('Jelmer Vernooij', 'jelmer@debian.org'))

        with open('debian/changelog') as f:
            self.assertEqual("""\
blah (2.0-1) UNRELEASED; urgency=low

  * New release

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 09 Nov 2020 15:05:05 -0000

lintian-brush (0.28) unstable; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", f.read())


class TextWrapperTests(TestCase):

    def setUp(self):
        super().setUp()
        self.wrapper = TextWrapper()

    def test_wrap_closes(self):
        self.assertEqual(
            ['And', ' ', 'this', ' ', 'fixes', ' ', 'something.', ' ',
             'Closes: #123456'],
            self.wrapper._split('And this fixes something. Closes: #123456'))


LONG_LINE = (
    "This is a very long line that could have been broken "
    "and should have been broken but was not broken.")


class RewrapChangeTests(TestCase):

    def test_too_short(self):
        self.assertEqual([], rewrap_change([]))
        self.assertEqual(['Foo bar'], rewrap_change(['Foo bar']))
        self.assertEqual(['Foo', 'bar'], rewrap_change(['Foo', 'bar']))
        self.assertEqual(
            ['  * Beginning', '  next line'],
            rewrap_change(
                ['  * Beginning', '  next line']))

    def test_no_initial(self):
        self.assertEqual(['x' * 100], rewrap_change(['x' * 100]))

    def test_wrap(self):
        self.assertEqual(
            ['  * This is a very long line that could have been '
             'broken and should have been',
             '    broken but was not broken.'],
            rewrap_change(['  * ' + LONG_LINE]))
        self.assertEqual("""\
  * Build-Depend on libsdl1.2-dev, libsdl-ttf2.0-dev and libsdl-mixer1.2-dev
    instead of with the embedded version, add -lSDL_ttf to --with-py-libs in
    debian/rules and rebootstrap (Closes: #382202)
""".splitlines(), rewrap_change("""\
  * Build-Depend on libsdl1.2-dev, libsdl-ttf2.0-dev and libsdl-mixer1.2-dev \
instead
    of with the embedded version, add -lSDL_ttf to --with-py-libs in \
debian/rules
    and rebootstrap (Closes: #382202)
""".splitlines()))

    def test_no_join(self):
        self.assertEqual("""\
    - Translators know why this sign has been put here:
        _Choices: ${FOO}, !Other[ You only have to translate Other, remove the
      exclamation mark and this comment between brackets]
      Currently text, newt, slang and gtk frontends support this feature.
""".splitlines(), rewrap_change("""\
    - Translators know why this sign has been put here:
        _Choices: ${FOO}, !Other[ You only have to translate Other, remove \
the exclamation mark and this comment between brackets]
      Currently text, newt, slang and gtk frontends support this feature.
""".splitlines()))


class IncVersionTests(TestCase):

    def test_native(self):
        self.assertEqual(Version('1.1'), increment_version(Version('1.0')))
        self.assertEqual(Version('1a1.1'), increment_version(Version('1a1.0')))
        self.assertEqual(
            Version('9.11~2'), increment_version(Version('9.11~1')))
        self.assertEqual(
            Version('9.11~1'), increment_version(Version('9.11~')))

    def test_non_native(self):
        self.assertEqual(Version('1.1-2'), increment_version(Version('1.1-1')))
        self.assertEqual(
            Version('9.11-1~1'), increment_version(Version('9.11-1~')))
        self.assertEqual(
            Version('9.11-1~2'), increment_version(Version('9.11-1~1')))


class ChangesSectionsTests(TestCase):

    def test_simple(self):
        self.assertEqual([
            (None, [1, 2, 3, 4], [
                ([(1, '  * Change 1')]),
                ([(2, '  * Change 2'), (3, '    rest')]),
            ])], list(changes_sections([
                '',
                '  * Change 1',
                '  * Change 2',
                '    rest',
                ''
                ])))

    def test_with_header(self):
        self.assertEqual([
            ('Author 1', [1, 2, 3], [([(2, '  * Change 1')])]),
            ('Author 2', [4, 5, 6, 7], [([(5, '  * Change 2'),
                                          (6, '    rest')])]),
            ], list(changes_sections([
                '',
                '  [ Author 1 ]',
                '  * Change 1',
                '',
                '  [ Author 2 ]',
                '  * Change 2',
                '    rest',
                '',
                ])))


class StripChangelogMessageTests(TestCase):

    def test_None(self):
        self.assertEqual(strip_changelog_message(None), None)  # type: ignore

    def test_no_changes(self):
        self.assertEqual(strip_changelog_message([]), [])

    def test_empty_changes(self):
        self.assertEqual(strip_changelog_message(['']), [])

    def test_removes_leading_whitespace(self):
        self.assertEqual(strip_changelog_message(
                    ['foo', '  bar', '\tbaz', '   bang']),
                    ['foo', 'bar', 'baz', ' bang'])

    def test_removes_star_if_one(self):
        self.assertEqual(strip_changelog_message(['  * foo']), ['foo'])
        self.assertEqual(strip_changelog_message(['\t* foo']), ['foo'])
        self.assertEqual(strip_changelog_message(['  + foo']), ['foo'])
        self.assertEqual(strip_changelog_message(['  - foo']), ['foo'])
        self.assertEqual(strip_changelog_message(['  *  foo']), ['foo'])
        self.assertEqual(
            strip_changelog_message(['  *  foo', '     bar']), ['foo', 'bar'])

    def test_leaves_start_if_multiple(self):
        self.assertEqual(
            strip_changelog_message(['  * foo', '  * bar']),
            ['* foo', '* bar'])
        self.assertEqual(
            strip_changelog_message(['  * foo', '  + bar']),
            ['* foo', '+ bar'])
        self.assertEqual(
            strip_changelog_message(
                ['  * foo', '  bar', '  * baz']),
            ['* foo', 'bar', '* baz'])


class TestNewUpstreamPackageVersion(TestCase):

    def test_simple_debian(self):
        self.assertEqual(
            Version("1.2-1"),
            new_upstream_package_version("1.2", "debian"))

    def test_simple_ubuntu(self):
        self.assertEqual(
            Version("1.2-0ubuntu1"),
            new_upstream_package_version("1.2", "ubuntu"))

    def test_debian_with_dash(self):
        self.assertEqual(
            Version("1.2-0ubuntu1-1"),
            new_upstream_package_version("1.2-0ubuntu1", "debian"))

    def test_ubuntu_with_dash(self):
        self.assertEqual(
            Version("1.2-1-0ubuntu1"),
            new_upstream_package_version("1.2-1", "ubuntu"))

    def test_ubuntu_with_epoch(self):
        self.assertEqual(
            Version("3:1.2-1-0ubuntu1"),
            new_upstream_package_version("1.2-1", "ubuntu", "3"))


class AllShaPrefixedTests(TestCase):

    def test_no_prefix(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        self.assertFalse(all_sha_prefixed(cl[0]))

    def test_empty(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        self.assertFalse(all_sha_prefixed(cl[0]))

    def test_prefixed(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        self.assertTrue(all_sha_prefixed(cl[0]))


class BlockEnsureFirstLineTests(TestCase):

    def test_ensure_first_line(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        changeblock_ensure_first_line(
            cl[0], 'QA Upload.',
            maintainer=('Jelmer Vernooij', 'jelmer@debian.org'))
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * QA Upload.
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))

    def test_before_author_block(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  [ Joe Example ]
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        changeblock_ensure_first_line(cl[0], 'QA Upload.')
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * QA Upload.

  [ Joe Example ]
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))

    def test_exists(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * QA Upload.
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        changeblock_ensure_first_line(cl[0], 'QA Upload.')
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * QA Upload.
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))

    def test_different_author(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        changeblock_ensure_first_line(
            cl[0], 'QA Upload.', maintainer=('Joe Example', 'joe@example.com'))
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * QA Upload.

  [ Jelmer Vernooij ]
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Joe Example <joe@example.com>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))


class ReleaseTests(TestCase):

    def test_already_updated(self):
        cl = Changelog("""\
lintian-brush (0.28) unstable; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        release(
                cl, timestamp=1604934305, localtime=False,
                maintainer=("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual(str(cl), """\
lintian-brush (0.28) unstable; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")

    def test_updated(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        release(cl, timestamp=1604934305, localtime=False,
                maintainer=("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual(str(cl), """\
lintian-brush (0.28) unstable; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 09 Nov 2020 15:05:05 -0000
""")

    def test_different_author(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Joe Example <joe@example.com>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        release(cl, timestamp=1604934305, localtime=False,
                maintainer=("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual(str(cl), """\
lintian-brush (0.28) unstable; urgency=medium

  [ Joe Example ]
  * [217263e] Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 09 Nov 2020 15:05:05 -0000
""")


class FindLastDistributionTests(TestCase):

    def create_changelog(self, *distributions):
        changelog = Changelog()
        changes = ["  [ A. Hacker ]", "  * Something"]
        author = "J. Maintainer <maint@example.com"
        for distro in distributions:
            changelog.new_block(changes=changes, author=author,
                                distributions=distro)
        return changelog

    def test_first(self):
        changelog = self.create_changelog("unstable")
        self.assertEqual("unstable", find_last_distribution(changelog))

    def test_second(self):
        changelog = self.create_changelog("unstable", "UNRELEASED")
        self.assertEqual("UNRELEASED", changelog.distributions)
        self.assertEqual("unstable", find_last_distribution(changelog))

    def test_empty(self):
        changelog = self.create_changelog()
        self.assertEqual(None, find_last_distribution(changelog))

    def test_only_unreleased(self):
        changelog = self.create_changelog("UNRELEASED")
        self.assertEqual(None, find_last_distribution(changelog))


class ChangelogInfoTests(TestCase):

    def test_find_extra_authors_none(self):
        changes = ["  * Do foo", "  * Do bar"]
        authors = find_extra_authors(changes)
        self.assertEqual([], authors)

    def test_find_extra_authors(self):
        changes = ["  * Do foo", "", "  [ A. Hacker ]", "  * Do bar", "",
                   "  [ B. Hacker ]", "  [ A. Hacker}"]
        authors = find_extra_authors(changes)
        self.assertEqual(["A. Hacker", "B. Hacker"], authors)
        self.assertEqual([str]*len(authors), list(map(type, authors)))

    def test_find_extra_authors_utf8(self):
        changes = ["  * Do foo", "", "  [ \xe1. Hacker ]", "  * Do bar", "",
                   "  [ \xe7. Hacker ]", "  [ A. Hacker}"]
        authors = find_extra_authors(changes)
        self.assertEqual(["\xe1. Hacker", "\xe7. Hacker"], authors)
        self.assertEqual([str]*len(authors), list(map(type, authors)))

    def test_find_extra_authors_iso_8859_1(self):
        # We try to treat lines as utf-8, but if that fails to decode, we fall
        # back to iso-8859-1
        changes = ["  * Do foo", "", "  [ \xe1. Hacker ]", "  * Do bar", "",
                   "  [ \xe7. Hacker ]", "  [ A. Hacker}"]
        authors = find_extra_authors(changes)
        self.assertEqual(["\xe1. Hacker", "\xe7. Hacker"], authors)
        self.assertEqual([str]*len(authors), list(map(type, authors)))

    def test_find_extra_authors_no_changes(self):
        authors = find_extra_authors([])
        self.assertEqual([], authors)

    def assert_thanks_is(self, changes, expected_thanks):
        thanks = find_thanks(changes)
        self.assertEqual(expected_thanks, thanks)
        self.assertEqual([str]*len(thanks), list(map(type, thanks)))

    def test_find_thanks_no_changes(self):
        self.assert_thanks_is([], [])

    def test_find_thanks_none(self):
        changes = ["  * Do foo", "  * Do bar"]
        self.assert_thanks_is(changes, [])

    def test_find_thanks(self):
        changes = ["  * Thanks to A. Hacker"]
        self.assert_thanks_is(changes, ["A. Hacker"])
        changes = ["  * Thanks to James A. Hacker"]
        self.assert_thanks_is(changes, ["James A. Hacker"])
        changes = ["  * Thankyou to B. Hacker"]
        self.assert_thanks_is(changes, ["B. Hacker"])
        changes = ["  * thanks to A. Hacker"]
        self.assert_thanks_is(changes, ["A. Hacker"])
        changes = ["  * thankyou to B. Hacker"]
        self.assert_thanks_is(changes, ["B. Hacker"])
        changes = ["  * Thanks A. Hacker"]
        self.assert_thanks_is(changes, ["A. Hacker"])
        changes = ["  * Thankyou B.  Hacker"]
        self.assert_thanks_is(changes, ["B. Hacker"])
        changes = ["  * Thanks to Mark A. Super-Hacker"]
        self.assert_thanks_is(changes, ["Mark A. Super-Hacker"])
        changes = ["  * Thanks to A. Hacker <ahacker@example.com>"]
        self.assert_thanks_is(changes, ["A. Hacker <ahacker@example.com>"])
        changes = ["  * Thanks to Adeodato Sim\xc3\xb3"]
        self.assert_thanks_is(changes, ["Adeodato Sim\xc3\xb3"])
        changes = ["  * Thanks to \xc1deodato Sim\xc3\xb3"]
        self.assert_thanks_is(changes, ["\xc1deodato Sim\xc3\xb3"])


class UpstreamMergeChangelogLineTests(TestCase):

    def test_release(self):
        self.assertEqual(
            "New upstream release.",
            upstream_merge_changelog_line("1.0"))

    def test_bzr_snapshot(self):
        self.assertEqual(
            "New upstream snapshot.",
            upstream_merge_changelog_line("1.0+bzr3"))

    def test_git_snapshot(self):
        self.assertEqual(
            "New upstream snapshot.",
            upstream_merge_changelog_line("1.0~git20101212"))

    def test_plus(self):
        self.assertEqual(
            "New upstream release.",
            upstream_merge_changelog_line("1.0+dfsg1"))


class TakeUploadershipTests(TestCase):

    def test_already_uploader(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        take_uploadership(cl[0], ("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))

    def test_other_author(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  * Add fixer for obsolete-runtime-tests-restriction.

 -- Joe Example <joe@example.com>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        take_uploadership(cl[0], ("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  [ Joe Example ]
  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))

    def test_other_author_2(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  [ Jane Example ]
  * Add fixer for obsolete-runtime-tests-restriction.

 -- Joe Example <joe@example.com>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        take_uploadership(cl[0], ("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

  [ Jane Example ]
  * Add fixer for obsolete-runtime-tests-restriction.

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))

    def test_other_author_empty(self):
        cl = Changelog("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

 -- Joe Example <joe@example.com>  Mon, 02 Sep 2019 00:23:11 +0000
""")
        take_uploadership(cl[0], ("Jelmer Vernooij", "jelmer@debian.org"))
        self.assertEqual("""\
lintian-brush (0.28) UNRELEASED; urgency=medium

 -- Jelmer Vernooij <jelmer@debian.org>  Mon, 02 Sep 2019 00:23:11 +0000
""", str(cl))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.versions."""

from datetime import datetime

from debmutate.versions import (
    add_dfsg_suffix,
    git_snapshot_data_from_version,
    mangle_version_for_git,
    new_package_version,
    get_snapshot_revision,
    upstream_version_add_revision,
    debianize_upstream_version,
    matches_release,
    strip_dfsg_suffix,
    )

from debian.changelog import Version
from unittest import TestCase


class MangleVersionForGitTests(TestCase):

    def test_simple(self):
        self.assertEqual(mangle_version_for_git('1.1'), '1.1')
        self.assertEqual(mangle_version_for_git('1.1a'), '1.1a')

    def test_mangled(self):
        self.assertEqual(mangle_version_for_git('1.1~a'), '1.1_a')
        self.assertEqual(mangle_version_for_git('1:1.1'), '1%1.1')
        self.assertEqual(mangle_version_for_git('1.1.'), '1.1.#')
        self.assertEqual(mangle_version_for_git('1.1..a'), '1.1.#.a')
        self.assertEqual(mangle_version_for_git('1.1.lock'), '1.1.#lock')


class GitSnapshotDataFromVersionTests(TestCase):

    def test_not_git(self):
        self.assertEqual((None, None), git_snapshot_data_from_version('1.1'))

    def test_git(self):
        self.assertEqual(
            (None, '2020-01-01'),
            git_snapshot_data_from_version('1.1+git20200101'))

    def test_next(self):
        self.assertEqual(
            (None, '2020-01-02'),
            git_snapshot_data_from_version('1.1+next.20200102'))


class TestPackageVersion(TestCase):

    def test_simple_debian(self):
        self.assertEqual(
            Version("1.2-1"),
            new_package_version("1.2", "debian"))

    def test_simple_ubuntu(self):
        self.assertEqual(
            Version("1.2-0ubuntu1"),
            new_package_version("1.2", "ubuntu"))

    def test_debian_with_dash(self):
        self.assertEqual(
            Version("1.2-0ubuntu1-1"),
            new_package_version("1.2-0ubuntu1", "debian"))

    def test_ubuntu_with_dash(self):
        self.assertEqual(
            Version("1.2-1-0ubuntu1"),
            new_package_version("1.2-1", "ubuntu"))

    def test_ubuntu_with_epoch(self):
        self.assertEqual(
            Version("3:1.2-1-0ubuntu1"),
            new_package_version("1.2-1", "ubuntu", "3"))


class GetRevisionSnapshotTests(TestCase):

    def test_with_snapshot(self):
        self.assertEqual(("bzr", "30"), get_snapshot_revision("0.4.4~bzr30"))

    def test_with_snapshot_plus(self):
        self.assertEqual(("bzr", "30"), get_snapshot_revision("0.4.4+bzr30"))

    def test_without_snapshot(self):
        self.assertEqual(None, get_snapshot_revision("0.4.4"))

    def test_non_numeric_snapshot(self):
        self.assertEqual(
            None,
            get_snapshot_revision("0.4.4~bzra"))

    def test_with_svn_snapshot(self):
        self.assertEqual(
            ("svn", "4242"), get_snapshot_revision("0.4.4~svn4242"))

    def test_with_svn_snapshot_plus(self):
        self.assertEqual(
            ("svn", "2424"), get_snapshot_revision("0.4.4+svn2424"))

    def test_git(self):
        self.assertEqual(
            ("date", "20190101"),
            get_snapshot_revision("0.4.4+git20190101"))
        self.assertEqual(
            ("git", "abc1def"),
            get_snapshot_revision("0.4.4+git20190101.abc1def"))


class TestUpstreamVersionAddRevision(TestCase):
    """Test that updating the version string works."""

    def setUp(self):
        super().setUp()
        self.revnos = {}
        self.svn_revnos = {b"somesvnrev": 45}
        self.git_shas = {
            b"somegitrev": b"e7f47cf254a6ddd4996fe41fa6115bd32eff5437"}
        self.revnos = {b"somerev": 42, b"somesvnrev": 12, b"somegitrev": 66}
        self.repository = self

    def revision_id_to_revno(self, revid):
        return self.revnos[revid]

    def revision_id_to_dotted_revno(self, revid):
        return (self.revnos[revid], )

    def test_update_plus_rev(self):
        self.assertEqual(
            "1.3+bzr42",
            upstream_version_add_revision("1.3+bzr23", bzr_revno="42"))

    def test_update_tilde_rev(self):
        self.assertEqual(
            "1.3~bzr42",
            upstream_version_add_revision("1.3~bzr23", bzr_revno="42"))

    def test_new_rev(self):
        self.assertEqual(
            "1.3+bzr42",
            upstream_version_add_revision("1.3", bzr_revno="42"))

    def test_svn_new_rev(self):
        self.assertEqual(
            "1.3+svn45",
            upstream_version_add_revision("1.3", svn_revno=45))

    def test_svn_plus_rev(self):
        self.assertEqual(
            "1.3+svn45",
            upstream_version_add_revision("1.3+svn3", svn_revno=45))

    def test_svn_tilde_rev(self):
        self.assertEqual(
            "1.3~svn45",
            upstream_version_add_revision("1.3~svn800", svn_revno=45))

    def test_git_tilde_rev(self):
        self.assertEqual(
            "1.3~git20180101.e7f47cf",
            upstream_version_add_revision(
                "1.3~git20171201.11b1d57",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))

    def test_git_new_rev(self):
        self.assertEqual(
            "1.3+git20180101.1.e7f47cf",
            upstream_version_add_revision(
                "1.3",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))
        self.assertEqual(
            "1.0~git20180101",
            upstream_version_add_revision(
                "1.0~git20160320",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))
        self.assertEqual(
            "1.0-git20180101",
            upstream_version_add_revision(
                "1.0-git20160320",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))
        self.assertEqual(
            "1.0~git20180101.1.e7f47cf",
            upstream_version_add_revision(
                "1.0~git20180101.0.11b1d57",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))
        self.assertEqual(
            "1.0~git20180101.0.e7f47cf",
            upstream_version_add_revision(
                "1.0~git20170101.0.11b1d57",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))
        self.assertEqual(
            "0.0~git20180101.0.e7f47cf",
            upstream_version_add_revision(
                "0.0~git20161231.0.3435554",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))

    def test_dfsg(self):
        self.assertEqual(
            "1.3+git20180101.1.e7f47cf",
            upstream_version_add_revision(
                "1.3+dfsg",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))
        self.assertEqual(
            "1.3+git20180101.1.e7f47cf",
            upstream_version_add_revision(
                "1.3+dfsg4",
                gitid=b'e7f47cfeaae7f47cfeaae7f47cfeaae7f47cfeaa',
                gitdate=datetime(2018, 1, 1)))


class DebianizeUpstreamVersionTests(TestCase):

    def test_unchanged(self):
        self.assertEqual('1.0', debianize_upstream_version('1.0'))

    def test_changed(self):
        self.assertEqual('1.0~beta1', debianize_upstream_version('1.0-beta1'))
        self.assertEqual('1.0~rc1', debianize_upstream_version('1.0-rc1'))
        self.assertEqual('1.0~a1', debianize_upstream_version('1.0a1'))


class MatchesReleaseTests(TestCase):

    def test_matches(self):
        self.assertTrue(matches_release('1.0', '1.0'))
        self.assertTrue(matches_release('1.0+ds1', '1.0'))
        self.assertTrue(
            matches_release('1.14.3+dfsg+~0.15.3', '0.15.3'))


class AddDfsgSuffixTests(TestCase):

    def test_new(self):
        self.assertEqual('1.0+ds', add_dfsg_suffix('1.0'))

    def test_existing(self):
        self.assertEqual('1.0+ds', add_dfsg_suffix('1.0', '0.9'))
        self.assertEqual('1.0+ds', add_dfsg_suffix('1.0', '0.9+ds'))
        self.assertEqual('1.0+ds1', add_dfsg_suffix('1.0', '0.9+ds1'))
        self.assertEqual('1.0+dfsg1', add_dfsg_suffix('1.0', '0.9+dfsg2'))
        self.assertEqual('1.0+dfsg', add_dfsg_suffix('1.0', '0.9+dfsg'))


class StripDfsgSuffixTests(TestCase):

    def test_nothing(self):
        self.assertEqual('1.0', strip_dfsg_suffix('1.0'))
        self.assertEqual('1.0+bzr2', strip_dfsg_suffix('1.0+bzr2'))

    def test_strip(self):
        self.assertEqual('1.0', strip_dfsg_suffix('1.0+ds1'))
        self.assertEqual('1.0', strip_dfsg_suffix('1.0+dfsg2'))
